from pyspedas.vires.load import load


def data(trange=None,
         collection=None,
         measurements=None,
         models=None,
         sampling_step=None,
         auxiliaries=None,
         residuals=False):
    return load(trange=trange,
                collection=collection,
                measurements=measurements,
                models=models,
                sampling_step=sampling_step,
                auxiliaries=auxiliaries,
                residuals=residuals)



# Copyright 2017 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pydivide

import logging
import re
import os
from . import download_files_utilities as utils
from .file_regex import kp_regex, l2_regex
import numpy as np
import collections


def param_list(kp):
    '''
    Return a listing of all parameters present in the given
    insitu data dictionary/structure.

    Input:
        kp: insitu kp data structure/dictionary read from file(s)
    Output:
        ParamList: a list of all contained items and their indices.
    '''
    import pandas as pd
    index = 1
    param_list_ = []
    for base_tag in kp.keys():
        if isinstance(kp[base_tag], pd.DataFrame):
            for obs_tag in kp[base_tag].columns:
                param_list_.append("#%3d %s.%s" % (index, base_tag, obs_tag))
                index += 1
        elif isinstance(kp[base_tag], pd.Series):
            param_list_.append("#%3d %s" % (index, base_tag))
            index += 1
        elif isinstance(kp[base_tag], pd.Index):
            param_list_.append("#%3d %s" % (index, base_tag))
            index += 1
        else:
            logging.warning('*****WARNING*****')
            logging.warning('Returning INCOMPLETE Parameter List')
            logging.warning('Base tag neither DataFrame nor Series')
            logging.warning('Plese check read_insitu_file definition')

    return param_list_


# ---------------------------------------------------------------------


def param_range(kp, iuvs=None):
    '''
    Print the range of times and orbits for the provided insitu data.
    If iuvs data are also provided, return only orbit numbers for IUVS data.

    Caveats:
        At present, not configured to handle (real) IUVS data.
        Current configuration of procedure assumes IUVS has identical
            time information as in-situ.

    Input:
        kp: insitu kp data structure/dictionary
        iuvs: IUVS kp data strucure/dictionary
    Output:
        None: prints information to screen
    '''

    # First, the case where insitu data are provided
    print("The loaded insitu KP data set contains data between")
    print("   %s and %s" % (np.array(kp['TimeString'])[0], np.array(kp['TimeString'])[-1]))
    print("Equivalently, this corresponds to orbits")
    print("   %6d and %6d." % (np.array(kp['Orbit'])[0], np.array(kp['Orbit'])[-1]))

    #  Next, the case where IUVS data are provided
    iuvs_data = False
    iuvs_tags = ['CORONA_LO_HIGH', 'CORONA_LO_LIMB', 'CORONA_LO_DISK',
                 'CORONA_E_HIGH', 'CORONA_E_LIMB', 'CORONA_E_DISK',
                 'APOAPSE', 'PERIAPSE', 'STELLAR_OCC']
    if kp.keys() in iuvs_tags:
        print("The loaded IUVS KP data set contains data between orbits")
        print("   %6d and %6d." % (np.array(kp['Orbit'])[0], np.array(kp['Orbit'])[-1]))

    #  Finally, the case where both insitu and IUVS are provided
    if iuvs is not None:
        print("The loaded IUVS KP data set contains data between orbits")
        print("   %6d and %6d." % (np.array(iuvs['Orbit'])[0], np.array(iuvs['Orbit'])[-1]))
        insitu_min, insitu_max = (np.nanmin([kp['Orbit']]), np.nanmax([kp['Orbit']]))
        if np.nanmax([iuvs['Orbit']]) < insitu_min or np.nanmin([iuvs['Orbit']]) > insitu_max:
            print("*** WARNING ***")
            print("There is NO overlap between the supplied insitu and IUVS")
            print("  data structures.  We cannot guarantee your safety ")
            print("  should you attempt to display these IUVS data against")
            print("  these insitu-supplied emphemeris data.")
    return  # No information to return


# --------------------------------------------------------------------------


def range_select(kp, time=None, parameter=None, maximum=None, minimum=None):
    '''
    Returns a subset of the input data based on the provided time
    and/or parameter criteria.  If neither Time nor Parameter filter
    information is provided, then no subselection of data will occur.
    Any parameter used as a filtering criterion must be paired with
    either a maximum and/or a minimum value.  Open ended bounds must
    be indicated with either a value of 'None' or an empty string ('').

    Input:
        kp: insitu kp data structure/dictionary read from file(s)
        time: two-element time range must be either strings of format
            'yyyy-mm-ddThh:mm:ss' or integers (orbit numbers)
        parameter: Element of provided data structure/dictionary by
            which to filter data.  Parameter(s) must be either integer
            type (search by index) or string type (search by instrument
            name and observation type).  If multiple Parameters are used
            to filter the data, they must be provided as a list (mixing
            data types within a list is permitted).
        maximum: maximum value of Parameter on which to filter.  A value of
            None or '' will leave the Parameter filter unbounded above.
            The number of elements of Maximum *MUST* equal the number of
            elements of Parameter.
        minimum: minimum value of Parameter on which to filter.  A value of
            None or '' will leave the Parameter filter unbounded below.
            The number of elements of Minimum *MUST* equal the number of
            elements of Parameter.
    Output: a dictionary/structure containing the same elements as the provided
        one, but filtered according to the Time and Parameter options.

    ToDo: compartmentalize the filtering and/or argument checks.
    '''
    from datetime import datetime

    #  Initialize the filter_list
    filter_list = []

    # First, check the arguments
    if time is None and parameter is None:
        insufficient_input_range_select()
        print('Neither Time nor Parameter provided')
        return kp
    elif time is None:
        # Then only subset based on parameters
        # Need to check whether one or several Parameters given
        inst = []
        obs = []
        if isinstance(parameter, int) or isinstance(parameter, str):
            # First, verify that at least one bound exists
            if minimum is None and maximum is None:
                insufficient_input_range_select()
                print('No bounds set for parameter: %s' % parameter)
                return kp
            elif minimum is None:
                # Range only bounded above
                minimum = -np.Infinity
            elif maximum is None:
                # range only bounded below
                maximum = np.Infinity
            else:
                # Range bounded on both ends
                pass
            a, b = get_inst_obs_labels(kp, parameter)
            inst.append(a)
            obs.append(b)
            nparam = 1  # necc?
        elif type(parameter) is list:
            nparam = len(parameter)
            for param in parameter:
                a, b = get_inst_obs_labels(kp, param)
                inst.append(a)
                obs.append(b)
        else:
            print('*****ERROR*****')
            print('Cannot identify given parameter: %s' % parameter)
            print('Suggest using param_list(kp) to identify Parameter')
            print('by index or by name')
            print('Returning complete original data dictionary')
            return kp

    # Should I move this below the Time conditional and move
    # Baselining of Filter List to above time

    else:
        # Time has been provided as a filtering agent
        # Determine whether Time is provided as strings or orbits
        if len(time) != 2:
            if parameter is not None:
                print('*****WARNING*****')
                print('Time must be provided as a two-element list')
                print('of either strings (yyyy-mm-dd hh:mm:ss) ')
                print('or orbits.  Since a Parameter *was* provided,')
                print('I will filter on that, but ignore the time input.')
            else:
                # Cannot proceed with filtering
                insufficient_input_range_select()
                print('Time malformed must be either a string of format')
                print('yyyy-mm-ddThh:mm:ss or integer orbit)')
                print('and no Parameter criterion given')
        else:
            # We have a two-element Time list: parse it
            if not isinstance(type(time[0]), type(time[1])):
                if parameter is not None:
                    print('*****WARNING*****')
                    print('Both elements of time must be same type')
                    print('Only strings of format yyyy-mm-dd hh:mm:ss')
                    print('or integers (orbit numbers) are allowed.')
                    print('Ignoring time inputs; will filter ONLY')
                    print('on Parameter inputs.')
                else:
                    print('*****ERROR*****')
                    print('Both elements of Time must be same type')
                    print('Only Strings of format yyyy-mm-dd hh:mm:ss')
                    print('or integers (orbit numbers) are allowed.')
                    print('Returning original unchanged data dictionary')
                    return kp
            elif type(time[0]) is int:
                # Filter based on orbit number
                min_t = min(time)
                max_t = max(time)
                filter_list.append(kp['Orbit'] >= min_t)
                filter_list.append(kp['Orbit'] <= max_t)
            elif isinstance(time[0], str):
                # Filter acc to string dat, need to parse it
                time_dt = [datetime.strptime(i, '%Y-%m-%d %H:%M:%S') for i in time]
                min_dt = min(time_dt)
                max_dt = max(time_dt)
                kp_dt = [datetime.strptime(i, '%Y-%m-%dT%H:%M:%S')
                         for i in kp['TimeString']]
                delta_tmin = np.array([(i - min_dt).total_seconds() for i in kp_dt])
                delta_tmax = np.array([(i - max_dt).total_seconds() for i in kp_dt])
                filter_list.append(delta_tmin >= 0)
                filter_list.append(delta_tmax <= 0)
            else:
                # Time provided as other than string or Integer
                if parameter is not None:
                    print('*****WARNING*****')
                    print('Both elements of time must be same type')
                    print('Only strings of format yyyy-mm-dd hh:mm:ss')
                    print('or integers (orbit numbers) are allowed.')
                    print('Ignoring time inputs; will filter ONLY')
                    print('on Parameter inputs.')
                else:
                    print('*****ERROR*****')
                    print('Both elements of Time must be same type')
                    print('Only Strings of format yyyy-mm-dd hh:mm:ss')
                    print('or integers (orbit numbers) are allowed.')
                    print('Returning original unchanged data dictionary')
                    return kp
            # Now, we apply the Parameter selection
            inst = []
            obs = []
            if isinstance(parameter, int) or isinstance(parameter, str):
                # Then we have a single Parameter to filter on
                # Verify that bounds info exists
                if minimum is None and maximum is None:
                    insufficient_input_range_select()
                    print('No bounds set for parameter %s' % parameter)
                    print('Applying only Time filtering')
                    parameter = None
                elif minimum is None:
                    minimum = -np.Infinity  # Unbounded below
                elif maximum is None:
                    maximum = np.Infinity  # Unbounded above
                else:
                    pass  # Range fully bounded
                a, b = get_inst_obs_labels(kp, parameter)
                inst.append(a)
                obs.append(b)
                nparam = 1  # necessary?
            elif type(parameter) is list:
                if len(parameter) != len(minimum) or len(parameter) != len(maximum):
                    print('*****ERROR*****')
                    print('---range_select---')
                    print('Number of minima and maxima provided')
                    print('MUST match number of Parameters provided')
                    print('You provided %4d Parameters' % len(parameter))
                    print('             %4d minima' % len(minimum))
                    print('         and %4d maxima' % len(maximum))
                    print('Filtering only on Time')
                    parameter = None
                else:
                    nparam = len(parameter)
                    for param in parameter:
                        a, b = get_inst_obs_labels(kp, param)
                        inst.append(a)
                        obs.append(b)

    # Now, apply the filters
    if parameter is not None:
        inst_obs_minmax = list(zip(inst, obs, minimum, maximum))
        for inst, obs, min_obs, max_obs in inst_obs_minmax:
            filter_list.append(kp[inst][obs] >= min_obs)
            filter_list.append(kp[inst][obs] <= max_obs)

    # Filter list built, apply to data
    filter = np.all(filter_list, axis=0)
    new = {}
    for i in kp:
        temp = kp[i]
        new.update({i: temp[filter]})
    return new


# --------------------------------------------------------------------------


def insufficient_input_range_select():
    '''
    This error message is called if user calls range_select with
    inputs that result in neither a valid Time range nor a valid
    Parameter range capable of being determined

    ToDo: Is there a way to hide this from the help feature?
    '''
    print('*****ERROR*****')
    print('Either a time criterion with two values.')
    print('  or a parameter name with maximum and/or')
    print('  minimum values must be provided.')
    print('Returning the complete original data dictionary')


# --------------------------------------------------------------------------

def get_inst_obs_labels(kp, name):
    '''
    Given parameter input in either string or integer format,
    identify the instrument name and observation type for use
    in accessing the relevant part of the data structure
    E.g.: 'LPW.EWAVE_LOW_FREQ' would be returned as
          ['LPW', 'EWAVE_LOW_FREQ']

    Input:
        kp: insitu kp data structure/dictionary read from file(s)
        name: string identifying a parameter.
            (Indices must be converted to inst.obs strings before
             calling this routine)
    Output:
        inst (1st arg): instrument identifier
        obs (2nd arg): observation type identifier
    '''

    # Need to ensure name is a string at this stage
    name = ('%s' % name)

    # Now, split at the dot (if it exists)
    tags = name.split('.')
    # And consider the various possibilities...
    if len(tags) == 2:
        return tags
    elif len(tags) == 1:
        try:
            int(tags[0])
            return (find_param_from_index(kp, tags[0])).split('.')
        except:
            print('*****ERROR*****')
            print('%s is an invalid parameter' % name)
            print('If only one value is provided, it must be an integer')
            return
    else:
        print('*****ERROR*****')
        print('%s is not a valid parameter' % name)
        print('because it has %1d elements' % len(tags))
        print('Only 1 integer or string of form "a.b" are allowed.')
        print('Please use .param_list attribute to find valid parameters')
        return


def find_param_from_index(kp, index):
    '''
    Given an integer index, find the name of the parameter

    Input:
        kp: insitu kp data structure/dictionary read from file(s)
        index: the index of the desired parameter (integer type)
    Output:
        A string of form <instrument>.<observation>
        (e.g., LPW.EWAVE_LOW_FREQ)
    '''

    index = '#%3d' % int(index)
    plist = param_list(kp)
    found = False
    for i in plist:
        if re.search(index, i):
            return i[5:]  # clip the '#123 ' string
    if not found:
        print('*****ERROR*****')
        print('%s not a valid index.' % index)
        print('Use param_list to list options')
        return


def remove_inst_tag(df):
    '''
    Remove the leading part of the column name that includes the instrument
    identifier for use in creating the parameter names for the toolkit.

    Input:
        A DataFrame produced from the insitu KP data
    Output:
        A new set of column names
    '''

    newcol = []
    for i in df.columns:
        if len(i.split('.')) >= 2:
            j = i.split('.')
            newcol.append('.'.join(j[1:]))

    return newcol


def get_latest_files_from_date_range(date1, date2):
    from datetime import timedelta

    mvn_root_data_dir = utils.get_root_data_dir()
    maven_data_dir = os.path.join(mvn_root_data_dir, 'maven', 'data', 'sci', 'kp', 'insitu')

    # Each file starts at midnight, so lets cut off the hours and just pay attention to the days
    date1 = date1.replace(hour=0, minute=0, second=0)
    date2 = date2.replace(hour=0, minute=0, second=0) + timedelta(days=1)

    time_spanned = date2 - date1
    num_days = time_spanned.days

    filenames = []

    for i in range(num_days):
        current_date = date1 + timedelta(days=i)
        year = str(current_date.year)
        month = str('%02d' % current_date.month)
        day = str('%02d' % current_date.day)
        full_path = os.path.join(maven_data_dir, year, month)
        if os.path.exists(full_path):
            # Grab only the most recent version/revision of regular and crustal insitu files for each
            # day
            insitu = {}
            c_insitu = {}
            for f in os.listdir(full_path):
                # print(f)
                if kp_regex.match(f).group('day') == day and not kp_regex.match(f).group('description'):
                    v = kp_regex.match(f).group('version')
                    r = kp_regex.match(f).group('revision')
                    insitu[f] = [v, r]
                elif kp_regex.match(f).group('day') == day and kp_regex.match(f).group('description') == '_crustal':
                    v = kp_regex.match(f).group('version')
                    r = kp_regex.match(f).group('revision')
                    c_insitu[f] = [v, r]

            if insitu:
                # Get max version
                insitu_file = max(insitu.keys(), key=(lambda k: insitu[k][0]))
                max_v = re.search('v\d{2}', insitu_file).group(0)
                # Get max revision
                max_r = max([re.search('r\d{2}', k).group(0) for k in insitu if max_v in k])
                # Get most recent insitu file (based on max version, and then max revision values)
                most_recent_insitu = [f for f in insitu.keys() if max_r in f and max_v in f]
                filenames.append(os.path.join(full_path, most_recent_insitu[0]))

            if c_insitu:
                # Get max version
                c_insitu_file = max(c_insitu.keys(), key=(lambda k: c_insitu[k][0]))
                c_max_v = re.search('v\d{2}', c_insitu_file).group(0)
                # Get max revision
                c_max_r = max([re.search('r\d{2}', k).group(0) for k in c_insitu if c_max_v in k])
                # Get most recent insitu file (based on max version, and then max revision values)
                most_recent_c_insitu = [f for f in c_insitu.keys() if c_max_r in f and c_max_v in f]
                filenames.append(os.path.join(full_path, most_recent_c_insitu[0]))

    filenames = sorted(filenames)
    return filenames


def get_latest_iuvs_files_from_date_range(date1, date2):
    from datetime import timedelta

    mvn_root_data_dir = utils.get_root_data_dir()
    maven_data_dir = os.path.join(mvn_root_data_dir, 'maven', 'data', 'sci', 'kp', 'iuvs')

    # Each file starts at midnight, so lets cut off the hours and just pay attention to the days
    date1 = date1.replace(hour=0, minute=0, second=0)
    date2 = date2.replace(day=date2.day, hour=0, minute=0, second=0) + timedelta(days=1)

    time_spanned = date2 - date1
    num_days = time_spanned.days

    files_to_return = []
    for i in range(num_days):
        current_date = date1 + timedelta(days=i)
        year = str(current_date.year)
        month = str('%02d' % current_date.month)
        day = str('%02d' % current_date.day)
        full_path = os.path.join(maven_data_dir, year, month)
        if os.path.exists(full_path):
            basenames = []
            # Obtain a list of all the basenames for the day
            for f in os.listdir(full_path):
                if kp_regex.match(f).group('day') == day:
                    description = kp_regex.match(f).group('description')
                    year = kp_regex.match(f).group('year')
                    month = kp_regex.match(f).group('month')
                    day = kp_regex.match(f).group('day')
                    time = kp_regex.match(f).group('time')
                    seq = ('mvn', 'kp', 'iuvs' + description, year + month + day + time)
                    basenames.append('_'.join(seq))

            basenames = list(set(basenames))

            for bn in basenames:
                version = 0
                revision = 0
                for f in os.listdir(full_path):
                    description = kp_regex.match(f).group('description')
                    year = kp_regex.match(f).group('year')
                    month = kp_regex.match(f).group('month')
                    day = kp_regex.match(f).group('day')
                    time = kp_regex.match(f).group('time')
                    seq = ('mvn', 'kp', 'iuvs' + description, year + month + day + time)
                    if bn == '_'.join(seq):
                        v = kp_regex.match(f).group('version')
                        if int(v) > int(version):
                            version = v

                for f in os.listdir(full_path):
                    description = kp_regex.match(f).group('description')
                    year = kp_regex.match(f).group('year')
                    month = kp_regex.match(f).group('month')
                    day = kp_regex.match(f).group('day')
                    time = kp_regex.match(f).group('time')
                    file_version = kp_regex.match(f).group('version')
                    seq = ('mvn', 'kp', 'iuvs' + description, year + month + day + time)
                    if bn == '_'.join(seq) and file_version == version:
                        r = kp_regex.match(f).group('revision')
                        if int(r) > int(revision):
                            revision = r
                if int(version) > 0:
                    seq = (bn, 'v' + str(version), 'r' + str(revision) + '.tab')
                    files_to_return.append(os.path.join(full_path, '_'.join(seq)))
    files_to_return = sorted(files_to_return)
    return files_to_return


def get_l2_files_from_date(date1, instrument):
    mvn_root_data_dir = utils.get_root_data_dir()
    maven_data_dir = os.path.join(mvn_root_data_dir, 'maven', 'data', 'sci', instrument, 'l2')

    # Each file starts at midnight, so lets cut off the hours and just pay attention to the days
    date1 = date1.replace(hour=0, minute=0, second=0)

    filenames = []

    year = str(date1.year)
    month = str('%02d' % date1.month)
    day = str('%02d' % date1.day)
    full_path = os.path.join(maven_data_dir, year, month)
    if os.path.exists(full_path):
        for f in os.listdir(full_path):
            if l2_regex.match(f).group('day') == day:
                filenames.append(os.path.join(full_path, f))

    filenames = sorted(filenames)
    return filenames


def get_header_info(filename):
    # Determine number of header lines
    nheader = 0
    with open(filename) as f:
        for line in f:
            if line.startswith('#'):
                nheader += 1

    # Parse the header (still needs special case work)
    read_param_list = False
    start_temp = False
    index_list = []
    with open(filename) as fin:
        icol = -2  # Counting header lines detailing column names
        iname = 1  # for counting seven lines with name info
        ncol = -1  # Dummy value to allow reading of early headerlines?
        col_regex = '#\s(.{16}){%3d}' % ncol  # needed for column names
        crustal = False
        if 'crustal' in filename:
            crustal = True
        for iline in range(nheader):
            line = fin.readline()
            # Define the proper indices change depending on the file type and row
            i = [2, 2, 1] if crustal else [1, 1, 1]
            if re.search('Number of parameter columns', line):
                ncol = int(re.split("\s{3}", line)[i[0]])
                # needed for column names
                col_regex = '#\s(.{16}){%2d}' % ncol if crustal else '#\s(.{16}){%3d}' % ncol
            elif re.search('Line on which data begins', line):
                nhead_test = int(re.split("\s{3}", line)[i[1]]) - 1
            elif re.search('Number of lines', line):
                ndata = int(re.split("\s{3}", line)[i[2]])
            elif re.search('PARAMETER', line):
                read_param_list = True
                param_head = iline
            elif read_param_list:
                icol += 1
                if icol > ncol:
                    read_param_list = False
            elif re.match(col_regex, line):
                # OK, verified match now get the values
                temp = re.findall('(.{16})', line[3:])
                if temp[0] == '               1':
                    start_temp = True
                if start_temp:
                    # Crustal files do not have as much variable info as other insitu files, need
                    # to modify the lines below
                    if crustal:
                        if iname == 1:
                            index = temp
                        elif iname == 2:
                            obs1 = temp
                        elif iname == 3:
                            obs2 = temp
                        elif iname == 4:
                            unit = temp
                            # crustal files don't come with this field
                            # throwing it in here for consistency with other insitu files
                            inst = ['     MODELED_MAG'] * 13
                        else:
                            print('More lines in data descriptor than expected.')
                            print('Line %d' % iline)
                    else:
                        if iname == 1:
                            index = temp
                        elif iname == 2:
                            obs1 = temp
                        elif iname == 3:
                            obs2 = temp
                        elif iname == 4:
                            obs3 = temp
                        elif iname == 5:
                            inst = temp
                        elif iname == 6:
                            unit = temp
                        elif iname == 7:
                            format_code = temp
                        else:
                            print('More lines in data descriptor than expected.')
                            print('Line %d' % iline)
                    iname += 1
            else:
                pass

        # Generate the names list.
        # NB, there are special case redundancies in there
        # (e.g., LPW: Electron Density Quality (min and max))
        # ****SWEA FLUX electron QUALITY *****
        first = True
        parallel = None
        names = []
        if crustal:
            for h, i, j in zip(inst, obs1, obs2):
                combo_name = (' '.join([i.strip(), j.strip()])).strip()
                # Add inst to names to avoid ambiguity
                # Will need to remove these after splitting
                names.append('.'.join([h.strip(), combo_name]))
                names[0] = 'Time'
        else:
            for h, i, j, k in zip(inst, obs1, obs2, obs3):
                combo_name = (' '.join([i.strip(), j.strip(), k.strip()])).strip()
                if re.match('^LPW$', h.strip()):
                    # Max and min error bars use same name in column
                    # SIS says first entry is min and second is max
                    if re.match('(Electron|Spacecraft)(.+)Quality', combo_name):
                        if first:
                            combo_name = combo_name + ' Min'
                            first = False
                        else:
                            combo_name = combo_name + ' Max'
                            first = True
                elif re.match('^SWEA$', h.strip()):
                    # electron flux qual flags do not indicate whether parallel or anti
                    # From context it is clear; but we need to specify in name
                    if re.match('.+Parallel.+', combo_name):
                        parallel = True
                    elif re.match('.+Anti-par', combo_name):
                        parallel = False
                    else:
                        pass
                    if re.match('Flux, e-(.+)Quality', combo_name):
                        if parallel:
                            p = re.compile('Flux, e- ')
                            combo_name = p.sub('Flux, e- Parallel ', combo_name)
                        else:
                            p = re.compile('Flux, e- ')
                            combo_name = p.sub('Flux, e- Anti-par ', combo_name)
                    if re.match('Electron eflux (.+)Quality', combo_name):
                        if parallel:
                            p = re.compile('Electron eflux ')
                            combo_name = p.sub('Electron eflux  Parallel ', combo_name)
                        else:
                            p = re.compile('Electron eflux ')
                            combo_name = p.sub('Electron eflux  Anti-par ', combo_name)
                # Add inst to names to avoid ambiguity
                # Will need to remove these after splitting
                names.append('.'.join([h.strip(), combo_name]))
                names[0] = 'Time'

    return names, inst


def initialize_list(the_list):
    index = 0
    for i in the_list:
        if hasattr(i, "__len__"):
            the_list[index] = initialize_list(i)
        else:
            the_list[index] = []
        index += 1
    return the_list


def place_values_in_list(the_list, location, to_append):
    testing = the_list
    if hasattr(location, "__len__"):
        for i in range(len(location)):
            testing = testing[location[i]]
        testing.append(to_append)
    else:
        testing = testing[location]
        testing.append(to_append)


def get_values_from_list(the_list, location):
    testing = the_list
    if hasattr(location, "__len__"):
        for i in range(len(location)):
            testing = testing[location[i]]
        return testing
    else:
        testing = testing[location]
        return testing


def orbit_time(begin_orbit, end_orbit=None):
    orb_file = os.path.join(os.path.dirname(__file__),
                            'maven_orb_rec.orb')

    with open(orb_file, "r") as f:
        if end_orbit is None:
            end_orbit = begin_orbit
        orbit_num = []
        time = []
        f.readline()
        f.readline()
        for line in f:
            line = line[0:28]
            line = line.split(' ')
            line = [x for x in line if x != '']
            orbit_num.append(int(line[0]))
            time.append(line[1] + "-" + month_to_num(line[2]) + "-" + line[3] + "T" + line[4])
        try:
            if orbit_num.index(begin_orbit) > len(time) or orbit_num.index(end_orbit) + 1 > len(time):
                print("Orbit numbers not found.  Please choose a number between 1 and %s.", orbit_num[-1])
                return [None, None]
            else:
                begin_time = time[orbit_num.index(begin_orbit)]
                end_time = time[orbit_num.index(end_orbit) + 1]
        except ValueError:
            return [None, None]
    return [begin_time, end_time]


def month_to_num(month_string):
    if month_string == 'JAN':
        return '01'
    if month_string == 'FEB':
        return '02'
    if month_string == 'MAR':
        return '03'
    if month_string == 'APR':
        return '04'
    if month_string == 'MAY':
        return '05'
    if month_string == 'JUN':
        return '06'
    if month_string == 'JUL':
        return '07'
    if month_string == 'AUG':
        return '08'
    if month_string == 'SEP':
        return '09'
    if month_string == 'OCT':
        return '10'
    if month_string == 'NOV':
        return '11'
    if month_string == 'DEC':
        return '12'


def mvn_kp_sc_traj_xyz(dims_x, dims_y, dims_z, values, x_array, y_array, z_array, nn='linear'):
    data = []
    if nn == 'nearest':
        for x, y, z in np.array([a for a in zip(x_array, y_array, z_array)]):
            ix = np.abs(dims_x - x).argmin()
            iy = np.abs(dims_y - y).argmin()
            iz = np.abs(dims_z - z).argmin()
            data.append(values[ix, iy, iz])
    else:
        max_x = np.max(x_array)
        min_x = np.min(x_array)
        max_y = np.max(y_array)
        min_y = np.min(y_array)
        max_z = np.max(z_array)
        min_z = np.min(z_array)

        for x, y, z in np.array([a for a in zip(x_array, y_array, z_array)]):
            if x > max_x:
                data.append(np.NaN)
            elif x < min_x:
                data.append(np.NaN)
            elif y > max_y:
                data.append(np.NaN)
            elif y < min_y:
                data.append(np.NaN)
            elif z > max_z:
                data.append(np.NaN)
            elif z < min_z:
                data.append(np.NaN)

            sorted_x_distance = np.argsort(np.abs(dims_x - x))
            ix1 = dims_x[sorted_x_distance[0]]
            ix2 = dims_x[sorted_x_distance[1]]
            if ix2 < ix1:
                temp = ix2
                ix2 = ix1
                ix1 = temp
            sorted_y_distance = np.argsort(np.abs(dims_y - y))
            iy1 = dims_y[sorted_y_distance[0]]
            iy2 = dims_y[sorted_y_distance[1]]
            if iy2 < iy1:
                temp = iy2
                iy2 = iy1
                iy1 = temp
            sorted_z_distance = np.argsort(np.abs(dims_z - z))
            iz1 = dims_z[sorted_z_distance[0]]
            iz2 = dims_z[sorted_z_distance[1]]
            if iz2 < iz1:
                temp = iz2
                iz2 = iz1
                iz1 = temp

            nx = (x - ix1) / (ix2 - ix1)
            ny = (y - iy1) / (iy2 - iy1)
            nz = (z - iz1) / (iz2 - iz1)

            data.append(values[sorted_x_distance[0], sorted_y_distance[0], sorted_z_distance[0]]
                        * (1 - nx) * (1 - ny) * (1 - nz) +
                        values[sorted_x_distance[1], sorted_y_distance[0], sorted_z_distance[0]]
                        * nx * (1 - ny) * (1 - nz) +
                        values[sorted_x_distance[0], sorted_y_distance[1], sorted_z_distance[0]]
                        * (1 - nx) * ny * (1 - nz) +
                        values[sorted_x_distance[0], sorted_y_distance[0], sorted_z_distance[1]]
                        * (1 - nx) * (1 - ny) * nz +
                        values[sorted_x_distance[1], sorted_y_distance[0], sorted_z_distance[1]]
                        * nx * (1 - ny) * nz +
                        values[sorted_x_distance[0], sorted_y_distance[1], sorted_z_distance[1]]
                        * (1 - nx) * ny * nz +
                        values[sorted_x_distance[1], sorted_y_distance[1], sorted_z_distance[0]]
                        * nx * ny * (1 - nz) +
                        values[sorted_x_distance[1], sorted_y_distance[1], sorted_z_distance[1]]
                        * nx * ny * nz)
    return data


def read_iuvs_file(file):
    iuvs_dict = {}
    periapse_num = 0
    occ_num = 0
    with open(file) as f:
        line = f.readline()
        while line != '':
            if line.startswith('*'):
                # Read the header
                line = f.readline()
                obs_mode = line[19:len(line) - 1].strip()

                header = {}
                f.readline()
                line = f.readline()
                header['time_start'] = line[19:len(line) - 1].strip()
                line = f.readline()
                header['time_stop'] = line[19:len(line) - 1].strip()
                line = f.readline()
                if obs_mode == "OCCULTATION":
                    header['target_name'] = line[19:len(line) - 1].strip()
                    line = f.readline()
                header['sza'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['local_time'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['lat'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['lon'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['lat_mso'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['lon_mso'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['orbit_number'] = int(line[19:len(line) - 1].strip())
                line = f.readline()
                header['mars_season'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sc_geo_x'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sc_geo_y'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sc_geo_z'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sc_mso_x'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sc_mso_y'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sc_mso_z'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sun_geo_x'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sun_geo_y'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sun_geo_z'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sun_geo_lat'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sun_geo_lon'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sun_mso_lat'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sun_mso_lon'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['subsol_geo_lon'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['subsol_geo_lat'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sc_sza'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sc_local_time'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['sc_alt'] = float(line[19:len(line) - 1].strip())
                line = f.readline()
                header['mars_sun_dist'] = float(line[19:len(line) - 1].strip())

                if obs_mode == "PERIAPSE":
                    periapse_num += 1
                    line = f.readline()
                    n_alt_bins = int(line[19:len(line) - 1].strip())
                    header['n_alt_bins'] = float(n_alt_bins)
                    line = f.readline()
                    n_alt_den_bins = int(line[19:len(line) - 1].strip())
                    header['n_alt_den_bins'] = float(n_alt_den_bins)

                    iuvs_dict['periapse' + str(periapse_num)] = {}
                    iuvs_dict['periapse' + str(periapse_num)].update(header)

                    # Empty space
                    f.readline()

                    # Read the Temperature
                    line = f.readline()
                    temp_labels = line[19:len(line) - 1].strip().split()
                    temperature = collections.OrderedDict((x, []) for x in temp_labels)
                    temperature_unc = collections.OrderedDict((x, []) for x in temp_labels)
                    line = f.readline()
                    vals = line[20:len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        temperature[list(temperature.keys())[index]].append(val)
                        index += 1
                    line = f.readline()
                    vals = line[20:len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        temperature_unc[list(temperature_unc.keys())[index]].append(val)
                        index += 1
                    iuvs_dict['periapse' + str(periapse_num)]['temperature'] = temperature
                    iuvs_dict['periapse' + str(periapse_num)]['temperature_unc'] = temperature_unc

                    # Empty space
                    f.readline()

                    # Read the Scale Heights
                    line = f.readline()
                    scale_height_labels = line[19:len(line) - 1].strip().split()
                    scale_height = collections.OrderedDict((x, []) for x in scale_height_labels)
                    scale_height_unc = collections.OrderedDict((x, []) for x in scale_height_labels)
                    line = f.readline()
                    vals = line[20:len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        scale_height[list(scale_height.keys())[index]].append(val)
                        index += 1
                    line = f.readline()
                    vals = line[20:len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        scale_height_unc[list(scale_height_unc.keys())[index]].append(val)
                        index += 1

                    iuvs_dict['periapse' + str(periapse_num)]['scale_height'] = scale_height
                    iuvs_dict['periapse' + str(periapse_num)]['scale_height_unc'] = scale_height_unc

                    # Empty space
                    f.readline()
                    f.readline()

                    # Read in the density
                    line = f.readline()
                    density_labels = line.strip().split()
                    density = collections.OrderedDict((x, []) for x in density_labels)
                    for i in range(0, n_alt_den_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == '-9.9999990E+09':
                                val = float('nan')
                            else:
                                val = float(val)
                            density[list(density.keys())[index]].append(val)
                            index += 1
                    iuvs_dict['periapse' + str(periapse_num)]['density'] = density

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the density systematic uncertainty
                    density_sys_unc = collections.OrderedDict((x, []) for x in density_labels)
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        density_sys_unc[list(density.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict['periapse' + str(periapse_num)]['density_sys_unc'] = density_sys_unc

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the density uncertainty
                    density_unc = collections.OrderedDict((x, []) for x in density_labels)
                    for i in range(0, n_alt_den_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == '-9.9999990E+09':
                                val = float('nan')
                            else:
                                val = float(val)
                            density_unc[list(density.keys())[index]].append(val)
                            index += 1
                    iuvs_dict['periapse' + str(periapse_num)]['density_sys_unc'] = density_sys_unc

                    f.readline()
                    f.readline()

                    line = f.readline()
                    radiance_labels = line.strip().split()
                    if "Cameron" in radiance_labels:
                        radiance_labels.remove('Cameron')
                    radiance = collections.OrderedDict((x, []) for x in radiance_labels)
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == '-9.9999990E+09':
                                val = float('nan')
                            else:
                                val = float(val)
                            radiance[list(radiance.keys())[index]].append(val)
                            index += 1

                    iuvs_dict['periapse' + str(periapse_num)]['radiance'] = radiance

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the radiance systematic uncertainty
                    radiance_sys_unc = collections.OrderedDict((x, []) for x in radiance_labels)
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        radiance_sys_unc[list(radiance.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict['periapse' + str(periapse_num)]['radiance_sys_unc'] = radiance_sys_unc

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the radiance uncertainty
                    radiance_unc = collections.OrderedDict((x, []) for x in radiance_labels)
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == '-9.9999990E+09':
                                val = float('nan')
                            else:
                                val = float(val)
                            radiance_unc[list(radiance.keys())[index]].append(val)
                            index += 1

                    iuvs_dict['periapse' + str(periapse_num)]['radiance_unc'] = radiance_unc

                elif obs_mode == "OCCULTATION":
                    occ_num += 1
                    line = f.readline()
                    n_alt_den_bins = int(line[19:len(line) - 1].strip())
                    header['n_alt_den_bins'] = float(n_alt_den_bins)

                    iuvs_dict['occultation' + str(occ_num)] = {}
                    iuvs_dict['occultation' + str(occ_num)].update(header)

                    # Empty space
                    f.readline()

                    # Read the Scale Heights
                    line = f.readline()
                    scale_height_labels = line[19:len(line) - 1].strip().split()
                    scale_height = collections.OrderedDict((x, []) for x in scale_height_labels)
                    scale_height_unc = collections.OrderedDict((x, []) for x in scale_height_labels)
                    line = f.readline()
                    vals = line[20:len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        scale_height[list(scale_height.keys())[index]].append(val)
                        index += 1
                    line = f.readline()
                    vals = line[20:len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        scale_height_unc[list(scale_height_unc.keys())[index]].append(val)
                        index += 1

                    iuvs_dict['occultation' + str(occ_num)]['scale_height'] = scale_height
                    iuvs_dict['occultation' + str(occ_num)]['scale_height_unc'] = scale_height_unc

                    # Empty space
                    f.readline()
                    f.readline()

                    # Read in the retrieval
                    line = f.readline()
                    retrieval_labels = line.strip().split()
                    retrieval = collections.OrderedDict((x, []) for x in retrieval_labels)
                    for i in range(0, n_alt_den_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == '-9.9999990E+09':
                                val = float('nan')
                            else:
                                val = float(val)
                            retrieval[list(retrieval.keys())[index]].append(val)
                            index += 1
                    iuvs_dict['occultation' + str(occ_num)]['retrieval'] = retrieval

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the retrieval systematic uncertainty
                    retrieval_sys_unc = collections.OrderedDict((x, []) for x in retrieval_labels)
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        retrieval_sys_unc[list(retrieval.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict['occultation' + str(occ_num)]['retrieval_sys_unc'] = retrieval_sys_unc

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the retrieval uncertainty
                    retrieval_unc = collections.OrderedDict((x, []) for x in retrieval_labels)
                    for i in range(0, n_alt_den_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == '-9.9999990E+09':
                                val = float('nan')
                            else:
                                val = float(val)
                            retrieval_unc[list(retrieval.keys())[index]].append(val)
                            index += 1
                    iuvs_dict['occultation' + str(occ_num)]['retrieval_sys_unc'] = retrieval_sys_unc

                elif obs_mode == "CORONA_LORES_HIGH":
                    line = f.readline()
                    n_alt_bins = int(line[19:len(line) - 1].strip())
                    header['n_alt_bins'] = float(n_alt_bins)

                    iuvs_dict['corona_lores_high'] = {}
                    iuvs_dict['corona_lores_high'].update(header)

                    f.readline()

                    # Read the Half int
                    line = f.readline()
                    half_int_dist_labels = line[19:len(line) - 1].strip().split()
                    half_int_dist = collections.OrderedDict((x, []) for x in half_int_dist_labels)
                    half_int_dist_unc = collections.OrderedDict((x, []) for x in half_int_dist_labels)
                    line = f.readline()
                    vals = line[26:len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        half_int_dist[list(half_int_dist.keys())[index]].append(val)
                        index += 1
                    line = f.readline()
                    vals = line[26:len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        half_int_dist_unc[list(half_int_dist_unc.keys())[index]].append(val)
                        index += 1

                    iuvs_dict['corona_lores_high']['half_int_dist'] = half_int_dist
                    iuvs_dict['corona_lores_high']['half_int_dist_unc'] = half_int_dist_unc

                    # Blank space
                    f.readline()
                    f.readline()

                    # Read in the density
                    line = f.readline()
                    density_labels = line.strip().split()
                    density = collections.OrderedDict((x, []) for x in density_labels)
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == '-9.9999990E+09':
                                val = float('nan')
                            else:
                                val = float(val)
                            density[list(density.keys())[index]].append(val)
                            index += 1

                    iuvs_dict['corona_lores_high']['density'] = density

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the density systematic uncertainty
                    density_sys_unc = collections.OrderedDict((x, []) for x in density_labels)
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        density_sys_unc[list(density.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict['corona_lores_high']['density_sys_unc'] = density_sys_unc

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the density uncertainty
                    density_unc = collections.OrderedDict((x, []) for x in density_labels)
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == '-9.9999990E+09':
                                val = float('nan')
                            else:
                                val = float(val)
                            density_unc[list(density.keys())[index]].append(val)
                            index += 1

                    iuvs_dict['corona_lores_high']['density_unc'] = density_unc

                    f.readline()
                    f.readline()

                    line = f.readline()
                    radiance_labels = line.strip().split()
                    if "Cameron" in radiance_labels:
                        radiance_labels.remove('Cameron')
                    radiance = collections.OrderedDict((x, []) for x in radiance_labels)
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == '-9.9999990E+09':
                                val = float('nan')
                            else:
                                val = float(val)
                            radiance[list(radiance.keys())[index]].append(val)
                            index += 1

                    iuvs_dict['corona_lores_high']['radiance'] = radiance

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the radiance systematic uncertainty
                    radiance_sys_unc = collections.OrderedDict((x, []) for x in radiance_labels)
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        radiance_sys_unc[list(radiance.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict['corona_lores_high']['radiance_sys_unc'] = radiance_sys_unc

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the radiance uncertainty
                    radiance_unc = collections.OrderedDict((x, []) for x in radiance_labels)
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == '-9.9999990E+09':
                                val = float('nan')
                            else:
                                val = float(val)
                            radiance_unc[list(radiance.keys())[index]].append(val)
                            index += 1

                    iuvs_dict['corona_lores_high']['radiance_unc'] = radiance_unc

                elif obs_mode == 'APOAPSE':

                    f.readline()
                    maps = {}
                    for j in range(0, 17):
                        var = f.readline().strip()
                        line = f.readline()
                        lons = line.strip().split()
                        lons = [float(x) for x in lons]
                        lats = []
                        data = []
                        for k in range(0, 45):
                            line = f.readline().strip().split()
                            lats.append(float(line[0]))
                            line_data = line[1:]
                            line_data = [float(x) if x != '-9.9999990E+09' else float('nan') for x in line_data]
                            data.append(line_data)

                        maps[var] = data
                        f.readline()

                    maps['latitude'] = lats
                    maps['longitude'] = lons

                    iuvs_dict['apoapse'] = {}
                    iuvs_dict['apoapse'].update(header)
                    iuvs_dict['apoapse'].update(maps)

                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the radiance systematic uncertainty
                    line = f.readline()
                    radiance_labels = line.strip().split()
                    radiance_sys_unc = collections.OrderedDict((x, []) for x in radiance_labels)
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == '-9.9999990E+09':
                            val = float('nan')
                        else:
                            val = float(val)
                        radiance_sys_unc[list(radiance.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict['apoapse']['radiance_sys_unc'] = radiance_sys_unc

            line = f.readline()

    return iuvs_dict


param_dict = {'Electron Density': 'ELECTRON_DENSITY',
              'Electron Density Quality Min': 'ELECTRON_DENSITY_QUAL_MIN',
              'Electron Density Quality Max': 'ELECTRON_DENSITY_QUAL_MAX',
              'Electron Temperature': 'ELECTRON_TEMPERATURE',
              'Electron Temperature Quality Min': 'ELECTRON_TEMPERATURE_QUAL_MIN',
              'Electron Temperature Quality Max': 'ELECTRON_TEMPERATURE_QUAL_MAX',
              'Spacecraft Potential': 'SPACECRAFT_POTENTIAL',
              'Spacecraft Potential Quality Min': 'SPACECRAFT_POTENTIAL_QUAL_MIN',
              'Spacecraft Potential Quality Max': 'SPACECRAFT_POTENTIAL_QUAL_MAX',
              'E-field Power 2-100 Hz': 'EWAVE_LOW_FREQ',
              'E-field 2-100 Hz Quality': 'EWAVE_LOW_FREQ_QUAL_QUAL',
              'E-field Power 100-800 Hz': 'EWAVE_MID_FREQ',
              'E-field 100-800 Hz Quality': 'EWAVE_MID_FREQ_QUAL_QUAL',
              'E-field Power 0.8-1.0 Mhz': 'EWAVE_HIGH_FREQ',
              'E-field 0.8-1.0 Mhz Quality': 'EWAVE_HIGH_FREQ_QUAL_QUAL',
              'EUV Irradiance 0.1-7.0 nm': 'IRRADIANCE_LOW',
              'Irradiance 0.1-7.0 nm Quality': 'IRRADIANCE_LOW_QUAL',
              'EUV Irradiance 17-22 nm': 'IRRADIANCE_MID',
              'Irradiance 17-22 nm Quality': 'IRRADIANCE_MID_QUAL',
              'EUV Irradiance Lyman-alpha': 'IRRADIANCE_LYMAN',
              'Irradiance Lyman-alpha Quality': 'IRRADIANCE_LYMAN_QUAL',
              'Solar Wind Electron Density': 'SOLAR_WIND_ELECTRON_DENSITY',
              'Solar Wind E- Density Quality': 'SOLAR_WIND_ELECTRON_DENSITY_QUAL',
              'Solar Wind Electron Temperature': 'SOLAR_WIND_ELECTRON_TEMPERATURE',
              'Solar Wind E- Temperature Quality': 'SOLAR_WIND_ELECTRON_TEMPERATURE_QUAL',
              'Flux, e- Parallel (5-100 ev)': 'ELECTRON_PARALLEL_FLUX_LOW',
              'Flux, e- Parallel (5-100 ev) Quality': 'ELECTRON_PARALLEL_FLUX_LOW_QUAL',
              'Flux, e- Parallel (100-500 ev)': 'ELECTRON_PARALLEL_FLUX_MID',
              'Flux, e- Parallel (100-500 ev) Quality': 'ELECTRON_PARALLEL_FLUX_MID_QUAL',
              'Flux, e- Parallel (500-1000 ev)': 'ELECTRON_PARALLEL_FLUX_HIGH',
              'Flux, e- Parallel (500-1000 ev) Quality': 'ELECTRON_PARALLEL_FLUX_HIGH_QUAL',
              'Flux, e- Anti-par (5-100 ev)': 'ELECTRON_ANTI_PARALLEL_FLUX_LOW',
              'Flux, e- Anti-par (5-100 ev) Quality': 'ELECTRON_ANTI_PARALLEL_FLUX_LOW_QUAL',
              'Flux, e- Anti-par (100-500 ev)': 'ELECTRON_ANTI_PARALLEL_FLUX_MID',
              'Flux, e- Anti-par (100-500 ev) Quality': 'ELECTRON_ANTI_PARALLEL_FLUX_MID_QUAL',
              'Flux, e- Anti-par (500-1000 ev)': 'ELECTRON_ANTI_PARALLEL_FLUX_HIGH',
              'Flux, e- Anti-par (500-1000 ev) Quality': 'ELECTRON_ANTI_PARALLEL_FLUX_HIGH_QUAL',
              'Electron eflux Parallel (5-100 ev)': 'ELECTRON_PARALLEL_FLUX_LOW',
              'Electron eflux Parallel (5-100 ev) Quality': 'ELECTRON_PARALLEL_FLUX_LOW_QUAL',
              'Electron eflux Parallel (100-500 ev)': 'ELECTRON_PARALLEL_FLUX_MID',
              'Electron eflux Parallel (100-500 ev) Quality': 'ELECTRON_PARALLEL_FLUX_MID_QUAL',
              'Electron eflux Parallel (500-1000 ev)': 'ELECTRON_PARALLEL_FLUX_HIGH',
              'Electron eflux Parallel (500-1000 ev) Quality': 'ELECTRON_PARALLEL_FLUX_HIGH_QUAL',
              'Electron eflux Anti-par (5-100 ev)': 'ELECTRON_ANTI_PARALLEL_FLUX_LOW',
              'Electron eflux Anti-par (5-100 ev) Quality': 'ELECTRON_ANTI_PARALLEL_FLUX_LOW_QUAL',
              'Electron eflux Anti-par (100-500 ev)': 'ELECTRON_ANTI_PARALLEL_FLUX_MID',
              'Electron eflux Anti-par (100-500 ev) Quality': 'ELECTRON_ANTI_PARALLEL_FLUX_MID_QUAL',
              'Electron eflux Anti-par (500-1000 ev)': 'ELECTRON_ANTI_PARALLEL_FLUX_HIGH',
              'Electron eflux Anti-par (500-1000 ev) Quality': 'ELECTRON_ANTI_PARALLEL_FLUX_HIGH_QUAL',
              'Electron Spectrum Shape': 'ELECTRON_SPECTRUM_SHAPE_PARAMETER',
              'Spectrum Shape Quality': 'ELECTRON_SPECTRUM_SHAPE_PARAMETER_QUAL',
              'H+ Density': 'HPLUS_DENSITY',
              'H+ Density Quality': 'HPLUS_DENSITY_QUAL',
              'H+ Flow Velocity MSO X': 'HPLUS_FLOW_VELOCITY_MSO_X',
              'H+ Flow MSO X Quality': 'HPLUS_FLOW_VELOCITY_MSO_X_QUAL',
              'H+ Flow Velocity MSO Y': 'HPLUS_FLOW_VELOCITY_MSO_Y',
              'H+ Flow MSO Y Quality': 'HPLUS_FLOW_VELOCITY_MSO_Y_QUAL',
              'H+ Flow Velocity MSO Z': 'HPLUS_FLOW_VELOCITY_MSO_Z',
              'H+ Flow MSO Z Quality': 'HPLUS_FLOW_VELOCITY_MSO_Z_QUAL',
              'H+ Temperature': 'HPLUS_TEMPERATURE',
              'H+ Temperature Quality': 'HPLUS_TEMPERATURE_QUAL',
              'Solar Wind Dynamic Pressure': 'SOLAR_WIND_DYNAMIC_PRESSURE',
              'Solar Wind Pressure Quality': 'SOLAR_WIND_DYNAMIC_PRESSURE_QUAL',
              'STATIC Quality Flag': 'STATIC_QUALITY_FLAG',
              'O+ Density': 'OPLUS_DENSITY',
              'O+ Density Quality': 'OPLUS_DENSITY_QUAL',
              'O2+ Density': 'O2PLUS_DENSITY',
              'O2+ Density Quality': 'O2PLUS_DENSITY_QUAL',
              'O+ Temperature': 'OPLUS_TEMPERATURE',
              'O+ Temperature Quality': 'OPLUS_TEMPERATURE_QUAL',
              'O2+ Temperature': 'O2PLUS_TEMPERATURE',
              'O2+ Temperature Quality': 'O2PLUS_TEMPERATURE_QUAL',
              'O2+ Flow Velocity MAVEN_APP X': 'O2PLUS_FLOW_VELOCITY_MAVEN_APP_X',
              'O2+ Flow MAVEN_APP X Quality': 'O2PLUS_FLOW_VELOCITY_MAVEN_APP_X_QUAL',
              'O2+ Flow Velocity MAVEN_APP Y': 'O2PLUS_FLOW_VELOCITY_MAVEN_APP_Y',
              'O2+ Flow MAVEN_APP Y Quality': 'O2PLUS_FLOW_VELOCITY_MAVEN_APP_Y_QUAL',
              'O2+ Flow Velocity MAVEN_APP Z': 'O2PLUS_FLOW_VELOCITY_MAVEN_APP_Z',
              'O2+ Flow MAVEN_APP Z Quality': 'O2PLUS_FLOW_VELOCITY_MAVEN_APP_Z_QUAL',
              'O2+ Flow Velocity MSO X': 'O2PLUS_FLOW_VELOCITY_MSO_X',
              'O2+ Flow MSO X Quality': 'O2PLUS_FLOW_VELOCITY_MSO_X_QUAL',
              'O2+ Flow Velocity MSO Y': 'O2PLUS_FLOW_VELOCITY_MSO_Y',
              'O2+ Flow MSO Y Quality': 'O2PLUS_FLOW_VELOCITY_MSO_Y_QUAL',
              'O2+ Flow Velocity MSO Z': 'O2PLUS_FLOW_VELOCITY_MSO_Z',
              'O2+ Flow MSO Z Quality': 'O2PLUS_FLOW_VELOCITY_MSO_Z_QUAL',
              'H+ Omni Flux': 'HPLUS_OMNI_DIRECTIONAL_FLUX',
              'H+ Energy': 'HPLUS_CHARACTERISTIC_ENERGY',
              'H+ Energy Quality': 'HPLUS_CHARACTERISTIC_ENERGY_QUAL',
              'He++ Omni Flux': 'HEPLUS_OMNI_DIRECTIONAL_FLUX',
              'He++ Energy': 'HEPLUS_CHARACTERISTIC_ENERGY',
              'He++ Energy Quality': 'HEPLUS_CHARACTERISTIC_ENERGY_QUAL',
              'O+ Omni Flux': 'OPLUS_OMNI_DIRECTIONAL_FLUX',
              'O+ Energy': 'OPLUS_CHARACTERISTIC_ENERGY',
              'O+ Energy Quality': 'OPLUS_CHARACTERISTIC_ENERGY_QUAL',
              'O2+ Omni Flux': 'O2PLUS_OMNI_DIRECTIONAL_FLUX',
              'O2+ Energy': 'O2PLUS_CHARACTERISTIC_ENERGY',
              'O2+ Energy Quality': 'O2PLUS_CHARACTERISTIC_ENERGY_QUAL',
              'H+ Direction MSO X': 'HPLUS_CHARACTERISTIC_DIRECTION_MSO_X',
              'H+ Direction MSO Y': 'HPLUS_CHARACTERISTIC_DIRECTION_MSO_Y',
              'H+ Direction MSO Z': 'HPLUS_CHARACTERISTIC_DIRECTION_MSO_Z',
              'H+ Angular Width': 'HPLUS_CHARACTERISTIC_ANGULAR_WIDTH',
              'H+ Width Quality': 'HPLUS_CHARACTERISTIC_ANGULAR_WIDTH_QUAL',
              'Pickup Ion Direction MSO X': 'DOMINANT_PICKUP_ION_CHARACTERISTIC_DIRECTION_MSO_X',
              'Pickup Ion Direction MSO Y': 'DOMINANT_PICKUP_ION_CHARACTERISTIC_DIRECTION_MSO_Y',
              'Pickup Ion Direction MSO Z': 'DOMINANT_PICKUP_ION_CHARACTERISTIC_DIRECTION_MSO_Z',
              'Pickup Ion Angular Width': 'DOMINANT_PICKUP_ION_CHARACTERISTIC_ANGULAR_WIDTH',
              'Pickup Ion Width Quality': 'DOMINANT_PICKUP_ION_CHARACTERISTIC_ANGULAR_WIDTH_QUAL',
              'Ion Flux FOV 1 F': 'ION_ENERGY_FLUX__FOV_1_F',
              'Ion Flux FOV 1F Quality': 'ION_ENERGY_FLUX__FOV_1_F_QUAL',
              'Ion Flux FOV 1 R': 'ION_ENERGY_FLUX__FOV_1_R',
              'Ion Flux FOV 1R Quality': 'ION_ENERGY_FLUX__FOV_1_R_QUAL',
              'Ion Flux FOV 2 F': 'ION_ENERGY_FLUX__FOV_2_F',
              'Ion Flux FOV 2F Quality': 'ION_ENERGY_FLUX__FOV_2_F_QUAL',
              'Ion Flux FOV 2 R': 'ION_ENERGY_FLUX__FOV_2_R',
              'Ion Flux FOV 2R Quality': 'ION_ENERGY_FLUX__FOV_2_R_QUAL',
              'Electron Flux FOV 1 F': 'ELECTRON_ENERGY_FLUX___FOV_1_F',
              'Electron Flux FOV 1F Quality': 'ELECTRON_ENERGY_FLUX___FOV_1_F_QUAL',
              'Electron Flux FOV 1 R': 'ELECTRON_ENERGY_FLUX___FOV_1_R',
              'Electron Flux FOV 1R Quality': 'ELECTRON_ENERGY_FLUX___FOV_1_R_QUAL',
              'Electron Flux FOV 2 F': 'ELECTRON_ENERGY_FLUX___FOV_2_F',
              'Electron Flux FOV 2F Quality': 'ELECTRON_ENERGY_FLUX___FOV_2_F_QUAL',
              'Electron Flux FOV 2 R': 'ELECTRON_ENERGY_FLUX___FOV_2_R',
              'Electron Flux FOV 2R Quality': 'ELECTRON_ENERGY_FLUX___FOV_2_R_QUAL',
              'Look Direction 1-F MSO X': 'LOOK_DIRECTION_1_F_MSO_X',
              'Look Direction 1-F MSO Y': 'LOOK_DIRECTION_1_F_MSO_Y',
              'Look Direction 1-F MSO Z': 'LOOK_DIRECTION_1_F_MSO_Z',
              'Look Direction 1-R MSO X': 'LOOK_DIRECTION_1_R_MSO_X',
              'Look Direction 1-R MSO Y': 'LOOK_DIRECTION_1_R_MSO_Y',
              'Look Direction 1-R MSO Z': 'LOOK_DIRECTION_1_R_MSO_Z',
              'Look Direction 2-F MSO X': 'LOOK_DIRECTION_2_F_MSO_X',
              'Look Direction 2-F MSO Y': 'LOOK_DIRECTION_2_F_MSO_Y',
              'Look Direction 2-F MSO Z': 'LOOK_DIRECTION_2_F_MSO_Z',
              'Look Direction 2-R MSO X': 'LOOK_DIRECTION_2_R_MSO_X',
              'Look Direction 2-R MSO Y': 'LOOK_DIRECTION_2_R_MSO_Y',
              'Look Direction 2-R MSO Z': 'LOOK_DIRECTION_2_R_MSO_Z',
              'Magnetic Field MSO X': 'MSO_X',
              'Magnetic MSO X Quality': 'MSO_X_QUAL',
              'Magnetic Field MSO Y': 'MSO_Y',
              'Magnetic MSO Y Quality': 'MSO_Y_QUAL',
              'Magnetic Field MSO Z': 'MSO_Z',
              'Magnetic MSO Z Quality': 'MSO_Z_QUAL',
              'Magnetic Field GEO X': 'GEO_X',
              'Magnetic GEO X Quality': 'GEO_X_QUAL',
              'Magnetic Field GEO Y': 'GEO_Y',
              'Magnetic GEO Y Quality': 'GEO_Y_QUAL',
              'Magnetic Field GEO Z': 'GEO_Z',
              'Magnetic GEO Z Quality': 'GEO_Z_QUAL',
              'Magnetic Field RMS Dev': 'RMS_DEVIATION',
              'Magnetic RMS Quality': 'RMS_DEVIATION_QUAL',
              'Density He': 'HE_DENSITY',
              'Density He Precision': 'HE_DENSITY_PRECISION',
              'Density He Quality': 'HE_DENSITY_QUAL',
              'Density O': 'O_DENSITY',
              'Density O Precision': 'O_DENSITY_PRECISION',
              'Density O Quality': 'O_DENSITY_QUAL',
              'Density CO': 'CO_DENSITY',
              'Density CO Precision': 'CO_DENSITY_PRECISION',
              'Density CO Quality': 'CO_DENSITY_QUAL',
              'Density N2': 'N2_DENSITY',
              'Density N2 Precision': 'N2_DENSITY_PRECISION',
              'Density N2 Quality': 'N2_DENSITY_QUAL',
              'Density NO': 'NO_DENSITY',
              'Density NO Precision': 'NO_DENSITY_PRECISION',
              'Density NO Quality': 'NO_DENSITY_QUAL',
              'Density Ar': 'AR_DENSITY',
              'Density Ar Precision': 'AR_DENSITY_PRECISION',
              'Density Ar Quality': 'AR_DENSITY_QUAL',
              'Density CO2': 'CO2_DENSITY',
              'Density CO2 Precision': 'CO2_DENSITY_PRECISION',
              'Density CO2 Quality': 'CO2_DENSITY_QUAL',
              'Density 32+': 'O2PLUS_DENSITY',
              'Density 32+ Precision': 'O2PLUS_DENSITY_PRECISION',
              'Density 32+ Quality': 'O2PLUS_DENSITY_QUAL',
              'Density 44+': 'CO2PLUS_DENSITY',
              'Density 44+ Precision': 'CO2PLUS_DENSITY_PRECISION',
              'Density 44+ Quality': 'CO2PLUS_DENSITY_QUAL',
              'Density 30+': 'NOPLUS_DENSITY',
              'Density 30+ Precision': 'NOPLUS_DENSITY_PRECISION',
              'Density 30+ Quality': 'NOPLUS_DENSITY_QUAL',
              'Density 16+': 'OPLUS_DENSITY',
              'Density 16+ Precision': 'OPLUS_DENSITY_PRECISION',
              'Density 16+ Quality': 'OPLUS_DENSITY_QUAL',
              'Density 28+': 'CO2PLUS_N2PLUS_DENSITY',
              'Density 28+ Precision': 'CO2PLUS_N2PLUS_DENSITY_PRECISION',
              'Density 28+ Quality': 'CO2PLUS_N2PLUS_DENSITY_QUAL',
              'Density 12+': 'CPLUS_DENSITY',
              'Density 12+ Precision': 'CPLUS_DENSITY_PRECISION',
              'Density 12+ Quality': 'CPLUS_DENSITY_QUAL',
              'Density 17+': 'OHPLUS_DENSITY',
              'Density 17+ Precision': 'OHPLUS_DENSITY_PRECISION',
              'Density 17+ Quality': 'OHPLUS_DENSITY_QUAL',
              'Density 14+': 'NPLUS_DENSITY',
              'Density 14+ Precision': 'NPLUS_DENSITY_PRECISION',
              'Density 14+ Quality': 'NPLUS_DENSITY_QUAL',
              'APP Attitude GEO X': 'ATTITUDE_GEO_X',
              'APP Attitude GEO Y': 'ATTITUDE_GEO_Y',
              'APP Attitude GEO Z': 'ATTITUDE_GEO_Z',
              'APP Attitude MSO X': 'ATTITUDE_MSO_X',
              'APP Attitude MSO Y': 'ATTITUDE_MSO_Y',
              'APP Attitude MSO Z': 'ATTITUDE_MSO_Z',
              'Spacecraft GEO X': 'GEO_X',
              'Spacecraft GEO Y': 'GEO_Y',
              'Spacecraft GEO Z': 'GEO_Z',
              'Spacecraft MSO X': 'MSO_X',
              'Spacecraft MSO Y': 'MSO_Y',
              'Spacecraft MSO Z': 'MSO_Z',
              'Spacecraft GEO Longitude': 'SUB_SC_LONGITUDE',
              'Spacecraft GEO Latitude': 'SUB_SC_LATITUDE',
              'Spacecraft Solar Zenith Angle': 'SZA',
              'Spacecraft Local Time': 'LOCAL_TIME',
              'Spacecraft Altitude Aeroid': 'ALTITUDE',
              'Spacecraft Attitude GEO X': 'ATTITUDE_GEO_X',
              'Spacecraft Attitude GEO Y': 'ATTITUDE_GEO_Y',
              'Spacecraft Attitude GEO Z': 'ATTITUDE_GEO_Z',
              'Spacecraft Attitude MSO X': 'ATTITUDE_MSO_X',
              'Spacecraft Attitude MSO Y': 'ATTITUDE_MSO_Y',
              'Spacecraft Attitude MSO Z': 'ATTITUDE_MSO_Z',
              'Mars Season (Ls)': 'MARS_SEASON',
              'Mars-Sun Distance': 'MARS_SUN_DISTANCE',
              'Subsolar Point GEO Longitude': 'SUBSOLAR_POINT_GEO_LONGITUDE',
              'Subsolar Point GEO Latitude': 'SUBSOLAR_POINT_GEO_LATITUDE',
              'Sub-Mars Point on the Sun Longitude': 'SUBMARS_POINT_SOLAR_LONGITUDE',
              'Sub-Mars Point on the Sun Latitude': 'SUBMARS_POINT_SOLAR_LATITUDE',
              'Rot matrix MARS -> MSO Row 1, Col 1': 'T11',
              'Rot matrix MARS -> MSO Row 1, Col 2': 'T12',
              'Rot matrix MARS -> MSO Row 1, Col 3': 'T13',
              'Rot matrix MARS -> MSO Row 2, Col 1': 'T21',
              'Rot matrix MARS -> MSO Row 2, Col 2': 'T22',
              'Rot matrix MARS -> MSO Row 2, Col 3': 'T23',
              'Rot matrix MARS -> MSO Row 3, Col 1': 'T31',
              'Rot matrix MARS -> MSO Row 3, Col 2': 'T32',
              'Rot matrix MARS -> MSO Row 3, Col 3': 'T33',
              'Rot matrix SPCCRFT -> MSO Row 1, Col 1': 'SPACECRAFT_T11',
              'Rot matrix SPCCRFT -> MSO Row 1, Col 2': 'SPACECRAFT_T12',
              'Rot matrix SPCCRFT -> MSO Row 1, Col 3': 'SPACECRAFT_T13',
              'Rot matrix SPCCRFT -> MSO Row 2, Col 1': 'SPACECRAFT_T21',
              'Rot matrix SPCCRFT -> MSO Row 2, Col 2': 'SPACECRAFT_T22',
              'Rot matrix SPCCRFT -> MSO Row 2, Col 3': 'SPACECRAFT_T23',
              'Rot matrix SPCCRFT -> MSO Row 3, Col 1': 'SPACECRAFT_T31',
              'Rot matrix SPCCRFT -> MSO Row 3, Col 2': 'SPACECRAFT_T32',
              'Rot matrix SPCCRFT -> MSO Row 3, Col 3': 'SPACECRAFT_T33'}



import unittest

import numpy as np

from pytplot import data_exists
import pyspedas
from pyspedas import time_double
from pyspedas.geopack import tt89
from pyspedas.geopack import tt96
from pyspedas.geopack import tt01
from pyspedas.geopack import tts04
from pyspedas.geopack.get_tsy_params import get_tsy_params
from pyspedas.geopack.get_w_params import get_w
from pyspedas import tinterpol
from pytplot import join_vec, store_data, get_data

trange = ['2015-10-16', '2015-10-17']


def get_params(model, g_variables=None):
    support_trange = [time_double(trange[0])-60*60*24, 
                      time_double(trange[1])+60*60*24]
    pyspedas.kyoto.dst(trange=support_trange)
    pyspedas.omni.data(trange=trange)
    join_vec(['BX_GSE', 'BY_GSM', 'BZ_GSM'])
    if model == 't01' and g_variables is None:
        g_variables = [6.0, 10.0]
    else:
        if g_variables is not None:
            if not isinstance(g_variables, str) and not isinstance(g_variables, np.ndarray):
                g_variables = None
    return get_tsy_params('kyoto_dst',
                    'BX_GSE-BY_GSM-BZ_GSM_joined',
                    'proton_density',
                    'flow_speed',
                    model,
                    pressure_tvar='Pressure',
                    g_variables=g_variables,
                    speed=True)


class LoadTestCases(unittest.TestCase):
    def test_igrf(self):
        mec_vars = pyspedas.mms.mec(trange=trange)
        tt89('mms1_mec_r_gsm', igrf_only=True)
        self.assertTrue(data_exists('mms1_mec_r_gsm_bt89'))

    def test_tt89(self):
        mec_vars = pyspedas.mms.mec(trange=trange)
        tt89('mms1_mec_r_gsm')
        self.assertTrue(data_exists('mms1_mec_r_gsm_bt89'))

    def test_tt96(self):
        mec_vars = pyspedas.mms.mec(trange=trange)
        params = get_params('t96')
        tinterpol('mms1_mec_r_gsm', 'proton_density')
        tt96('mms1_mec_r_gsm-itrp', parmod=params)
        self.assertTrue(data_exists('mms1_mec_r_gsm-itrp_bt96'))

    def test_tt01(self):
        mec_vars = pyspedas.mms.mec(trange=trange)
        params = get_params('t01')
        tinterpol('mms1_mec_r_gsm', 'proton_density')
        tt01('mms1_mec_r_gsm-itrp', parmod=params)
        self.assertTrue(data_exists('mms1_mec_r_gsm-itrp_bt01'))
        mec = get_data('mms1_mec_r_gsm-itrp')
        gvars = np.zeros((len(mec.times), 2))
        gvars[:, 0] = np.repeat(6.0, len(mec.times))
        gvars[:, 1] = np.repeat(10.0, len(mec.times))
        store_data('g_variables', data={'x': mec.times, 'y': gvars})
        params = get_params('t01', g_variables='g_variables')
        tt01('mms1_mec_r_gsm-itrp', parmod=params)
        self.assertTrue(data_exists('mms1_mec_r_gsm-itrp_bt01'))
        params = get_params('t01', g_variables=gvars)
        tt01('mms1_mec_r_gsm-itrp', parmod=params)
        self.assertTrue(data_exists('mms1_mec_r_gsm-itrp_bt01'))

    def test_tts04(self):
        mec_vars = pyspedas.mms.mec(trange=trange)
        params = get_params('ts04')
        tinterpol('mms1_mec_r_gsm', 'proton_density')
        tts04('mms1_mec_r_gsm-itrp', parmod=params)
        self.assertTrue(data_exists('mms1_mec_r_gsm-itrp_bts04'))

    def test_get_w(self):
        w_vals = get_w(trange=['2015-10-16', '2015-10-17'])

    def test_errors(self):
        # exercise some of the error code
        mec_vars = pyspedas.mms.mec(trange=trange)
        params = get_params('ts04')
        tinterpol('mms1_mec_r_gsm', 'proton_density')
        tts04('var_doesnt_exist')
        tts04('mms1_mec_r_gsm-itrp', parmod=None)
        tt01('var_doesnt_exist')
        tt01('mms1_mec_r_gsm-itrp', parmod=None)
        tt96('var_doesnt_exist')
        tt96('mms1_mec_r_gsm-itrp', parmod=None)
        tt89('var_doesnt_exist')
        invalidmodel = get_params('89')
        invalidg = get_params('t01', g_variables=1)
        invalidg = get_params('t01', g_variables='g_vars')
        notrange = get_w()  # no trange
        invalidtrange = get_w(trange=['2050-01-01', '2050-01-02'])


if __name__ == '__main__':
    unittest.main()



import numpy as np


def slice2d_get_ebounds(dist):
    """
    Returns an array of gapless energy boundaries.  The number of
    elements returned will always be N+1 for N energy levels.

    Input
    ------
        dist: dict
            3D particle data structure

    Returns
    --------
        Array of energy bin boundaries (# energy bins + 1)
    """
    n = dist['energy'].shape[0]+1
    energies = np.zeros((n, dist['energy'][:, :, :].shape[1], dist['energy'][:, :, :].shape[2]))

    # use midpoints
    energies[1:n-1, :, :] = (dist['energy'][0:n-2, :, :] + dist['energy'][1:n-1, :, :])/2.0

    # top/bottom energies
    energies[0, :, :] = dist['energy'][0, :, :] + (dist['energy'][0, :, :] - energies[1, :, :])
    energies[n-1, :, :] = dist['energy'][n-2, :, :] + (dist['energy'][n-2, :, :] - energies[n-2, :, :])

    return energies



import logging
import numpy as np

from pytplot import get_data

from pyspedas.utilities.interpol import interpol
from pyspedas.particles.spd_part_products.spd_pgs_make_tplot import spd_pgs_make_tplot
from pyspedas.particles.spd_part_products.spd_pgs_limit_range import spd_pgs_limit_range
from pyspedas.particles.spd_part_products.spd_pgs_progress_update import spd_pgs_progress_update
from pyspedas.particles.spd_part_products.spd_pgs_do_fac import spd_pgs_do_fac
from pyspedas.particles.spd_part_products.spd_pgs_regrid import spd_pgs_regrid
from pyspedas.particles.moments.spd_pgs_moments import spd_pgs_moments
from pyspedas.particles.moments.spd_pgs_moments_tplot import spd_pgs_moments_tplot

from pyspedas.mms.fpi.mms_get_fpi_dist import mms_get_fpi_dist
from pyspedas.mms.hpca.mms_get_hpca_dist import mms_get_hpca_dist
from pyspedas.mms.particles.mms_convert_flux_units import mms_convert_flux_units
from pyspedas.mms.particles.mms_pgs_clean_data import mms_pgs_clean_data
from pyspedas.mms.particles.mms_pgs_clean_support import mms_pgs_clean_support
from pyspedas.mms.particles.mms_pgs_make_fac import mms_pgs_make_fac
from pyspedas.mms.particles.mms_pgs_split_hpca import mms_pgs_split_hpca
from pyspedas.mms.particles.mms_pgs_make_e_spec import mms_pgs_make_e_spec
from pyspedas.mms.particles.mms_pgs_make_phi_spec import mms_pgs_make_phi_spec
from pyspedas.mms.particles.mms_pgs_make_theta_spec import mms_pgs_make_theta_spec
from pyspedas.mms.particles.mms_part_des_photoelectrons import mms_part_des_photoelectrons

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def mms_part_products(in_tvarname,
                      units='eflux',
                      species='e',
                      data_rate='fast',
                      instrument='fpi',
                      probe='1',
                      output=['energy', 'theta', 'phi'],
                      energy=None,
                      phi=None,
                      theta=None,
                      pitch=None,
                      gyro=None,
                      mag_name=None,
                      pos_name=None,
                      fac_type='mphigeo',
                      sc_pot_name=None,
                      correct_photoelectrons=False,
                      zero_negative_values=False,
                      internal_photoelectron_corrections=False,
                      disable_photoelectron_corrections=False,
                      no_regrid=False,
                      regrid=[32, 16],
                      vel_name=None):
    """
    Generate spectra and moments from 3D MMS particle data; note: this routine isn't
    meant to be called directly - see the wrapper mms_part_getspec instead.

    Input
    ----------
        in_tvarname: str
            Name of the tplot variable containing MMS 3D particle distribution data

    Parameters
    ----------
        units: str
            Specify units of output variables; must be 'eflux' to calculate moments

            valid options:
            'flux'   -  # / (cm^2 * s * sr * eV)
            'eflux'  -  eV / (cm^2 * s * sr * eV)  <default>
            'df_cm'  -  s^3 / cm^6
            'df_km'  -  s^3 / km^6

        species: str
            Specify the species of the input tplot variable

        data_rate: str
            Data rate of the input data

        instrument: str
            Instrument (FPI or HPCA)

        probe: int or str
            Spacecraft probe #

        output: str or list of str
            Output variables; options: 
                'energy': energy spectrograms
                'theta': theta spectrograms
                'phi': phi spectrograms
                'pa': pitch-angle spectrograms
                'gyro': gyro-phase spectrograms
                'moments': plasma moments

        energy: list of float
            Energy range [min, max], in eV

        phi: list of float
            Phi range [min, max], in degrees

        theta: list of float
            Theta range [min, max], in degrees

        pitch: list of float
            Pitch-angle range [min, max], in degrees

        gyro: list of float
            Gyro-angle range [min, max], in degrees

        mag_name: str
            Tplot variable containing magnetic field data for
            moments and FAC transformations

        pos_name: str
            Tplot variable containing spacecraft position for
            FAC transformations

        sc_pot_name: str
            Tplot variable containing spacecraft potential data
            for moments corrections

        fac_type: str
            Field aligned coordinate system variant; default: 'mphigeo'
            options: 'phigeo', 'mphigeo', 'xgse'

        correct_photoelectrons: bool
            Flag to correct FPI data for photoelectrons 
            (defaults to True for FPI electron data - disable with the parameter below)

        disable_photoelectron_corrections: bool
            Flag to disable FPI photoelectron corrections

        internal_photoelectron_corrections: bool
            Apply internal photoelectron corrections

        zero_negative_values: bool
            Turn negative values to 0 after doing the photoelectron corrections (DES)
            
    Returns
    ----------
        Creates tplot variables containing spectrograms and moments

    """

    units = units.lower()
    if isinstance(probe, int):
        probe = str(probe)

    data_in = get_data(in_tvarname)

    if data_in is None:
        logging.error('Error, could not find the variable: ' + in_tvarname)
        return None

    if isinstance(output, str):
        output = output.split(' ')

    if instrument == 'fpi':
        dist_in = mms_get_fpi_dist(in_tvarname, species=species, probe=probe, data_rate=data_rate)
    elif instrument == 'hpca':
        dist_in = mms_get_hpca_dist(in_tvarname, species=species, probe=probe, data_rate=data_rate)
    else:
        logging.error('Error, unknown instrument: ' + instrument + '; valid options: fpi, hpca')
        return

    if instrument == 'hpca':
        data_times = mms_get_hpca_dist(in_tvarname, species=species, probe=probe, data_rate=data_rate, times=True)
    else:
        data_times = data_in.times

    # ntimes = len(data_times)
    ntimes = len(dist_in)

    # create rotation matrix to field aligned coordinates if needed
    fac_outputs = ['pa', 'gyro', 'fac_energy', 'fac_moments']
    fac_requested = len(set(output).intersection(fac_outputs)) > 0
    if fac_requested:
        fac_matrix = mms_pgs_make_fac(data_times, mag_name, pos_name, fac_type=fac_type)

        if fac_matrix is None:
            # problem creating the FAC matrices
            fac_requested = False

    out_energy = np.zeros((ntimes, dist_in[0]['n_energy']))
    out_energy_y = np.zeros((ntimes, dist_in[0]['n_energy']))
    out_theta = np.zeros((ntimes, dist_in[0]['n_theta']))
    out_phi = np.zeros((ntimes, dist_in[0]['n_phi']))
    out_theta_y = np.zeros((ntimes, dist_in[0]['n_theta']))
    out_phi_y = np.zeros((ntimes, dist_in[0]['n_phi']))
    if fac_requested:
        out_pad = np.zeros((ntimes, dist_in[0]['n_theta']))
        out_pad_y = np.zeros((ntimes, dist_in[0]['n_theta']))
        out_gyro = np.zeros((ntimes, dist_in[0]['n_phi']))
        out_gyro_y = np.zeros((ntimes, dist_in[0]['n_phi']))

    # moments
    if 'moments' in output:
        out_density = np.zeros(ntimes)
        out_avgtemp = np.zeros(ntimes)
        out_vthermal = np.zeros(ntimes)
        out_flux = np.zeros([ntimes, 3])
        out_velocity = np.zeros([ntimes, 3])
        out_mftens = np.zeros([ntimes, 6])
        out_ptens = np.zeros([ntimes, 6])
        #out_ttens = np.zeros([dist_in['n_times'], 3, 3])

    out_vars = []
    last_update_time = None

    if 'moments' in output or correct_photoelectrons or internal_photoelectron_corrections:
        support_data = mms_pgs_clean_support(data_times, mag_name=mag_name, vel_name=vel_name, sc_pot_name=sc_pot_name)
        mag_data = support_data[0]
        scpot_data = support_data[2]

    if disable_photoelectron_corrections:
        correct_photoelectrons = False
        internal_photoelectron_corrections = False

    # Remind user that l2 moments should be used preferentially and FAC moments are experimental
    if 'moments' in output or 'moments_fac' in output:
        msg = """Moments generated with mms_part_products may be missing several important
corrections, including photoelectron removal and spacecraft potential.
The official moments released by the instrument teams include these and
are the scientific products that should be used for analysis."""
        logging.warning('==================================================================================')
        logging.warning('WARNING:')
        logging.warning(msg)
        logging.warning('==================================================================================')

    # grab the DES photoelectron model if needed
    if (instrument != 'fpi' or species != 'e') and (correct_photoelectrons or internal_photoelectron_corrections):
        logging.error('Photoelectron corrections only valid for DES; no corrections will be applied.')
        correct_photoelectrons = False
        internal_photoelectron_corrections = False

    if correct_photoelectrons or internal_photoelectron_corrections:
        fpi_photoelectrons = mms_part_des_photoelectrons(in_tvarname)

        if fpi_photoelectrons is None:
            logging.error('Photoelectron model missing for this date; re-run without photoelectron corrections')
            return

        # will need stepper parities for burst mode data
        if data_rate == 'brst':
            parity = get_data('mms'+probe+'_des_steptable_parity_brst')

        startdelphi = get_data('mms'+probe+'_des_startdelphi_count_'+data_rate)

    for i in range(0, ntimes):
        last_update_time = spd_pgs_progress_update(last_update_time=last_update_time, current_sample=i, total_samples=ntimes, type_string=in_tvarname)

        if instrument == 'fpi':
            dists = mms_get_fpi_dist(in_tvarname, index=i, species=species, probe=probe, data_rate=data_rate)
        elif instrument == 'hpca':
            dists = mms_get_hpca_dist(in_tvarname, index=i, species=species, probe=probe, data_rate=data_rate)

        if isinstance(dists, list):
            dist_in = dists[0]
        else:
            dist_in = dists

        # apply the DES photoelectron corrections
        if correct_photoelectrons or internal_photoelectron_corrections:
            # From Dan Gershman's release notes on the FPI photoelectron model:
            # Find the index I in the startdelphi_counts_brst or startdelphi_counts_fast array
            # [360 possibilities] whose corresponding value is closest to th = e measured
            # startdelphi_count_brst or startdelphi_count_fast for the skymap of interest. The
            # closest index can be approximated by I = floor(startdelphi_count_brst/16) or I =
            # floor(startdelphi_count_fast/16)
            startdelphi_I = int(np.floor(startdelphi.y[i]/16.0))

            if data_rate == 'brst':
                parity_num = str(int(np.fix(parity.y[i])))

                bg_dist = fpi_photoelectrons['bgdist_p'+parity_num]
                n_value = fpi_photoelectrons['n_'+parity_num]

                fphoto = bg_dist.y[startdelphi_I, :, :, :]

                # need to interpolate using SC potential data to get Nphoto value
                nphoto_scpot_dependent = n_value.y[startdelphi_I, :]
                nphoto = interpol(nphoto_scpot_dependent, n_value.v, scpot_data[i])
            else:
                fphoto = fpi_photoelectrons['bg_dist'].y[startdelphi_I, :, :, :]

                # need to interpolate using SC potential data to get Nphoto value
                nphoto_scpot_dependent = fpi_photoelectrons['n'].y[startdelphi_I, :]
                nphoto = interpol(nphoto_scpot_dependent, fpi_photoelectrons['n'].v, scpot_data[i])

            # now, the corrected distribution function is simply f_corrected = f-fphoto*nphoto
            # note: transpose is to shuffle fphoto*nphoto to energy-azimuth-elevation, to match dist.data
            correction = fphoto*nphoto
            corrected_df = dist_in['data']-correction.transpose([2, 0, 1])

            if zero_negative_values:
                corrected_df[corrected_df < 0] = 0.0

            dist_in['data'] = corrected_df

        data = mms_convert_flux_units(dist_in, units=units)

        clean_data = mms_pgs_clean_data(data)

        # split hpca angle bins to be equal width in phi/theta
        # this is needed when skipping the regrid step
        if instrument == 'hpca':
            clean_data = mms_pgs_split_hpca(clean_data)

        # Apply phi, theta, & energy limits
        if energy is not None or theta is not None or phi is not None:
            clean_data = spd_pgs_limit_range(clean_data, energy=energy, theta=theta, phi=phi)

        # Build energy spectrogram
        if 'energy' in output:
            out_energy_y[i, :], out_energy[i, :] = mms_pgs_make_e_spec(clean_data)

        # Build theta spectrogram
        if 'theta' in output:
            out_theta_y[i, :], out_theta[i, :] = mms_pgs_make_theta_spec(clean_data, resolution=dist_in['n_theta'])

        # Build phi spectrogram
        if 'phi' in output:
            out_phi_y[i, :], out_phi[i, :] = mms_pgs_make_phi_spec(clean_data, resolution=dist_in['n_phi'])

        # Calculate the moments
        if 'moments' in output:
            if scpot_data is not None:
                scpot_val = scpot_data[i]
            else:
                scpot_val = 0.0

            moments = spd_pgs_moments(clean_data, sc_pot=scpot_val)
            out_density[i] = moments['density']
            out_avgtemp[i] = moments['avgtemp']
            out_vthermal[i] = moments['vthermal']
            out_flux[i, :] = moments['flux']
            out_velocity[i, :] = moments['velocity']
            out_mftens[i, :] = moments['mftens']
            out_ptens[i, :] = moments['ptens']

        # Perform transformation to FAC, regrid data, and apply limits in new coords
        if fac_requested:
            fac_data = spd_pgs_do_fac(clean_data, fac_matrix[i, :, :])

            if no_regrid == False:
                fac_data = spd_pgs_regrid(fac_data, regrid)

            fac_data['theta'] = 90.0-fac_data['theta']
            fac_data = spd_pgs_limit_range(fac_data, theta=pitch, phi=gyro)

        if 'pa' in output:
            out_pad_y[i, :], out_pad[i, :] = mms_pgs_make_theta_spec(fac_data, colatitude=True, resolution=dist_in['n_theta'])

        if 'gyro' in output:
            out_gyro_y[i, :], out_gyro[i, :] = mms_pgs_make_phi_spec(fac_data, resolution=dist_in['n_phi'])


    if 'moments' in output:
        # put all of the moments arrays into a hash table prior to passing to the tplot routine
        moments = {'density': out_density, 
              'flux': out_flux, 
              'mftens': out_mftens, 
              'velocity': out_velocity, 
              'ptens': out_ptens,
              'vthermal': out_vthermal,
              'avgtemp': out_avgtemp}
        moments_vars = spd_pgs_moments_tplot(moments, x=data_times, prefix=in_tvarname)
        out_vars.extend(moments_vars)

    if 'energy' in output:
        spd_pgs_make_tplot(in_tvarname+'_energy', x=data_times, y=out_energy_y, z=out_energy, units=units, ylog=True, ytitle=dist_in['data_name'], ysubtitle='energy (eV)')
        out_vars.append(in_tvarname+'_energy')

    if 'theta' in output:
        spd_pgs_make_tplot(in_tvarname+'_theta', x=data_times, y=out_theta_y, z=out_theta, units=units, ytitle=dist_in['data_name'], ysubtitle='theta (deg)')
        out_vars.append(in_tvarname+'_theta')

    if 'phi' in output:
        spd_pgs_make_tplot(in_tvarname+'_phi', x=data_times, y=out_phi_y, z=out_phi, units=units, ytitle=dist_in['data_name'], ysubtitle='phi (deg)')
        out_vars.append(in_tvarname+'_phi')

    if 'pa' in output:
        spd_pgs_make_tplot(in_tvarname+'_pa', x=data_times, y=out_pad_y, z=out_pad, units=units, ytitle=dist_in['data_name'], ysubtitle='PA (deg)')
        out_vars.append(in_tvarname+'_pa')

    if 'gyro' in output:
        spd_pgs_make_tplot(in_tvarname+'_gyro', x=data_times, y=out_gyro_y, z=out_gyro, units=units, ytitle=dist_in['data_name'], ysubtitle='gyro (deg)')
        out_vars.append(in_tvarname+'_gyro')

    return out_vars



from .load import load
from pyspedas.utilities.datasets import find_datasets


def vhm(trange=['2009-01-01', '2009-01-02'],
        datatype='1min', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=True):
    """
    This function loads data from the VHM/FGM experiment from the Ulysses mission
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='vhm', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def swoops(trange=['2009-01-01', '2009-01-02'],
        datatype='bai_m0', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the SWOOPS experiment from the Ulysses mission
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='swoops', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def swics(trange=['2009-01-01', '2009-01-02'],
        datatype='scs_m1', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the SWICS experiment from the Ulysses mission
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='swics', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def urap(trange=['2003-01-01', '2003-01-02'],
        datatype='pfrp_m0', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the URAP experiment from the Ulysses mission
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='urap', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def epac(trange=['1996-01-01', '1996-01-02'],
        datatype='epac_m1', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the EPAC experiment from the Ulysses mission
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='epac', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def hiscale(trange=['2003-01-01', '2003-01-02'],
        datatype='lmde_m1', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the HI-SCALE experiment from the Ulysses mission
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='hiscale', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def cospin(trange=['2003-01-01', '2003-01-02'],
        datatype='het', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the COSPIN experiment from the Ulysses mission
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='cospin', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def grb(trange=['2003-01-01', '2003-01-02'],
        datatype='grb_m0', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the GRB experiment from the Ulysses mission
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='grb', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)


def datasets(instrument=None, label=True):
    return find_datasets(mission='Ulysses', instrument=instrument, label=label)




import os
import unittest
from pytplot import data_exists

import pyspedas

class LoadTestCases(unittest.TestCase):
    def test_load_hep_data(self):
        hep_vars = pyspedas.erg.hep()
        self.assertTrue(data_exists('erg_hep_l2_FEDO_L'))
        self.assertTrue(data_exists('erg_hep_l2_FEDO_H'))

    def test_load_xep_data(self):
        xep_vars = pyspedas.erg.xep()
        self.assertTrue(data_exists('erg_xep_l2_FEDO_SSD'))

    def test_load_orb_data(self):
        orb_vars = pyspedas.erg.orb()
        self.assertTrue(data_exists('erg_orb_l2_pos_gse'))
        self.assertTrue(data_exists('erg_orb_l2_pos_gsm'))
        self.assertTrue(data_exists('erg_orb_l2_pos_sm'))
        self.assertTrue(data_exists('erg_orb_l2_vel_gse'))
        self.assertTrue(data_exists('erg_orb_l2_vel_gsm'))
        self.assertTrue(data_exists('erg_orb_l2_vel_sm'))

    def test_load_l3_orb_data(self):
        orb_vars = pyspedas.erg.orb(level='l3')
        self.assertTrue(data_exists('erg_orb_l3_pos_eq_op'))
        self.assertTrue(data_exists('erg_orb_l3_pos_iono_north_op'))
        self.assertTrue(data_exists('erg_orb_l3_pos_iono_south_op'))

    def test_load_mgf_data(self):
        mgf_vars = pyspedas.erg.mgf(time_clip=True)
        self.assertTrue(data_exists('erg_mgf_l2_mag_8sec_sm'))

    def test_load_lepe_data(self):
        lepe_vars = pyspedas.erg.lepe()
        self.assertTrue(data_exists('erg_lepe_l2_omniflux_FEDO'))

    def test_load_lepi_data(self):
        lepi_vars = pyspedas.erg.lepi()
        self.assertTrue(data_exists('erg_lepi_l2_omniflux_FPDO'))
        self.assertTrue(data_exists('erg_lepi_l2_omniflux_FHEDO'))
        self.assertTrue(data_exists('erg_lepi_l2_omniflux_FODO'))

    def test_load_mepe_data(self):
        mepe_vars = pyspedas.erg.mepe()
        self.assertTrue(data_exists('erg_mepe_l2_omniflux_FEDO'))

    def test_load_mepi_data(self):
        mepi_vars = pyspedas.erg.mepi_nml()
        self.assertTrue(data_exists('erg_mepi_l2_omniflux_epoch_tof'))
        mepi_vars = pyspedas.erg.mepi_tof()
        self.assertTrue(data_exists('erg_mepi_l2_tofflux_FPDU'))
        self.assertTrue(data_exists('erg_mepi_l2_tofflux_FODU'))

    def test_load_pwe_ofa_data(self):
        pwe_vars = pyspedas.erg.pwe_ofa()
        self.assertTrue(data_exists('erg_pwe_ofa_l2_spec_E_spectra_132'))
        self.assertTrue(data_exists('erg_pwe_ofa_l2_spec_B_spectra_132'))

    # test is failing as of 26 Jan 2022
    # def test_load_pwe_efd_data(self):
    #     pwe_vars = pyspedas.erg.pwe_efd()
    #     self.assertTrue(data_exists('erg_pwe_efd_l2_E_spin_Eu_dsi'))
    #     self.assertTrue(data_exists('erg_pwe_efd_l2_E_spin_Ev_dsi'))
    #     self.assertTrue(data_exists('erg_pwe_efd_l2_E_spin_Ev1_dsi'))
    #     self.assertTrue(data_exists('erg_pwe_efd_l2_E_spin_Eu2_dsi'))

    def test_load_pwe_hfa_data(self):
        pwe_vars = pyspedas.erg.pwe_hfa()
        self.assertTrue(data_exists('erg_pwe_hfa_l2_low_spectra_eu'))
        self.assertTrue(data_exists('erg_pwe_hfa_l2_low_spectra_ev'))
        self.assertTrue(data_exists('erg_pwe_hfa_l2_low_spectra_esum'))
        self.assertTrue(data_exists('erg_pwe_hfa_l2_low_spectra_er'))

    def test_downloadonly(self):
        files = pyspedas.erg.mgf(downloadonly=True, trange=['2017-03-27', '2017-03-28'])
        self.assertTrue(os.path.exists(files[0]))

if __name__ == '__main__':
    unittest.main()


import numpy as np
from pyspedas import tnames
from pytplot import get_data, store_data


def remove_duplicated_tframe(tvars=[]):

    tvars = tnames(tvars)

    if len(tvars) < 1:
        return

    for tvar in tvars:

        input_attr_dict = get_data(tvar, metadata=True)
        get_data_vars = get_data(tvar)

        unique_array, counts_array = np.unique(
            get_data_vars[0], return_counts=True)
        duplicate_time_indices = np.where(counts_array > 1)[0]
        if duplicate_time_indices.shape[0] > 0:  # duplication check
            delete_indices_array = np.array(
                [np.where(get_data_vars[0] == unique_array[index])[0][0]
                    for index in duplicate_time_indices])
            input_data_dictionary = {}
            input_data_dictionary['x'] = np.delete(
                get_data_vars[0], delete_indices_array, axis=0)
            input_data_dictionary['y'] = np.delete(
                get_data_vars[1], delete_indices_array, axis=0)

            # for v or v1, v2.. elements. not tested yet.
            if len(get_data_vars) >= 3:
                element_counts = len(get_data_vars)
                if element_counts == 3:
                    if get_data_vars[2].ndim >= 2:
                        input_data_dictionary['v'] = np.delete(
                            get_data_vars[2], delete_indices_array, axis=0)
                    elif get_data_vars[2].ndim == 1:
                        input_data_dictionary['v'] = get_data_vars[2]
                elif element_counts > 3:
                    for element_index in range(2, element_counts):
                        v_element_name = f'v{element_index-1}'
                        if get_data_vars[element_index].ndim >= 2:
                            input_data_dictionary[v_element_name] = np.delete(
                                get_data_vars[element_index], delete_indices_array,
                                axis=0)
                        elif get_data_vars[element_index].ndim == 1:
                            input_data_dictionary[v_element_name] = get_data_vars[element_index]

            store_data(tvar, data=input_data_dictionary,
                       attr_dict=input_attr_dict)

    return



import unittest
from pyspedas import mms_load_eis, mms_eis_pad
from pyspedas.mms.eis.mms_eis_omni import mms_eis_omni
from pytplot import data_exists
from pyspedas.mms.eis.mms_eis_spec_combine_sc import mms_eis_spec_combine_sc
from pytplot import tplot

class EISTestCases(unittest.TestCase):
    def test_electronenergy(self):
        mms_load_eis(datatype='electronenergy')
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_electronenergy_electron_flux_omni'))
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_electronenergy_electron_flux_omni_spin'))
        tplot('mms1_epd_eis_srvy_l2_electronenergy_electron_flux_omni', display=False)
        tplot('mms1_epd_eis_srvy_l2_electronenergy_electron_flux_omni_spin', display=False)

    def test_pad_extof_srvy(self):
        mms_load_eis(datatype='extof')
        mms_eis_pad(datatype='extof')
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_extof_46-10489keV_proton_flux_omni_pad_spin'))
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_extof_46-10489keV_proton_flux_omni_pad'))
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_extof_proton_flux_omni_spin'))
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_extof_proton_flux_omni'))
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_extof_oxygen_energy_range'))
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_extof_proton_energy_range'))
        tplot('mms1_epd_eis_srvy_l2_extof_proton_flux_omni_spin', display=False)
        tplot('mms1_epd_eis_srvy_l2_extof_46-10489keV_proton_flux_omni_pad_spin', display=False)
        tplot('mms1_epd_eis_srvy_l2_extof_46-10489keV_proton_flux_omni_pad', display=False)

    def test_pad_extof_srvy_probe(self):
        mms_load_eis(probe=4)
        mms_eis_omni(probe=4, data_units='cps')
        mms_eis_omni(probe=4, data_units='counts')
        mms_eis_pad(probe=4, data_units='cps')
        self.assertTrue(data_exists('mms4_epd_eis_srvy_l2_extof_44-1315keV_proton_cps_omni_pad_spin'))
        self.assertTrue(data_exists('mms4_epd_eis_srvy_l2_extof_44-1315keV_proton_cps_omni_pad'))
        self.assertTrue(data_exists('mms4_epd_eis_srvy_l2_extof_proton_cps_omni'))
        self.assertTrue(data_exists('mms4_epd_eis_srvy_l2_extof_oxygen_energy_range'))
        self.assertTrue(data_exists('mms4_epd_eis_srvy_l2_extof_proton_energy_range'))
        self.assertTrue(data_exists('mms4_epd_eis_srvy_l2_extof_proton_counts_omni'))
        tplot('mms4_epd_eis_srvy_l2_extof_proton_counts_omni', display=False)
        tplot('mms4_epd_eis_srvy_l2_extof_44-1315keV_proton_cps_omni_pad', display=False)
        tplot('mms4_epd_eis_srvy_l2_extof_44-1315keV_proton_cps_omni_pad_spin', display=False)

    def test_pad_extof_brst(self):
        mms_load_eis(probe=4, datatype='extof', data_rate='brst', trange=['2022-03-03/07:05:00', '2022-03-03/07:08:00'])
        mms_eis_pad(probe=4, datatype='extof', data_rate='brst')
        self.assertTrue(data_exists('mms4_epd_eis_brst_l2_extof_52-866keV_proton_flux_omni_pad_spin'))
        self.assertTrue(data_exists('mms4_epd_eis_brst_l2_extof_52-866keV_proton_flux_omni_pad'))
        self.assertTrue(data_exists('mms4_epd_eis_brst_l2_extof_proton_flux_omni'))
        self.assertTrue(data_exists('mms4_epd_eis_brst_l2_extof_proton_energy_range'))
        self.assertTrue(data_exists('mms4_epd_eis_brst_l2_extof_oxygen_energy_range'))
        tplot('mms4_epd_eis_brst_l2_extof_52-866keV_proton_flux_omni_pad', display=False)
        tplot('mms4_epd_eis_brst_l2_extof_52-866keV_proton_flux_omni_pad_spin', display=False)
    
    def test_load_phxtof_data(self):
        data = mms_load_eis(trange=['2015-10-16', '2015-10-16/01:00'], datatype='phxtof')
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_phxtof_proton_flux_omni'))
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_phxtof_proton_t5_energy_dminus'))
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_phxtof_proton_t5_energy_dplus'))
        tplot('mms1_epd_eis_srvy_l2_phxtof_proton_flux_omni', display=False)

    def test_load_phxtof_spdf(self):
        data = mms_load_eis(trange=['2015-10-16', '2015-10-16/01:00'], datatype='phxtof', spdf=True)
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_phxtof_proton_flux_omni'))
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_phxtof_proton_t5_energy_dminus'))
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_phxtof_proton_t5_energy_dplus'))
        tplot('mms1_epd_eis_srvy_l2_phxtof_proton_flux_omni', display=False)

    def test_load_extof_suffix(self):
        data = mms_load_eis(trange=['2015-10-16', '2015-10-17'], datatype='extof', suffix='_test')
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_extof_proton_flux_omni_test_spin'))
        self.assertTrue(data_exists('mms1_epd_eis_srvy_l2_extof_proton_flux_omni_test'))
        tplot('mms1_epd_eis_srvy_l2_extof_proton_flux_omni_test_spin', display=False)
        tplot('mms1_epd_eis_srvy_l2_extof_proton_flux_omni_test', display=False)

    def test_combine_sc(self):
        trange = ['2022-03-03/07:05:00', '2022-03-03/07:08:00']
        mms_load_eis(probe=[1, 2, 3, 4], datatype='extof', data_rate='brst', trange=trange)
        mms_load_eis(probe=[1, 2, 3, 4], datatype='phxtof', data_rate='brst', trange=trange)
        mms_eis_spec_combine_sc(datatype='extof', data_rate='brst')
        mms_eis_spec_combine_sc(datatype='phxtof', data_rate='brst')
        self.assertTrue(data_exists('mmsx_epd_eis_brst_l2_extof_proton_flux_omni'))
        self.assertTrue(data_exists('mmsx_epd_eis_brst_l2_extof_proton_flux_omni_spin'))
        self.assertTrue(data_exists('mmsx_epd_eis_brst_l2_phxtof_proton_flux_omni'))
        self.assertTrue(data_exists('mmsx_epd_eis_brst_l2_phxtof_proton_flux_omni_spin'))
        tplot('mmsx_epd_eis_brst_l2_extof_proton_flux_omni', display=False)
        tplot('mmsx_epd_eis_brst_l2_extof_proton_flux_omni_spin', display=False)
        tplot('mmsx_epd_eis_brst_l2_phxtof_proton_flux_omni', display=False)
        tplot('mmsx_epd_eis_brst_l2_phxtof_proton_flux_omni_spin', display=False)


if __name__ == '__main__':
    unittest.main()



'''
Load data from SOSMAG Space Weather instrument
    flying on the Korean GEO-KOMPSAT-2A satellite.

This function downloads data directly from the ESA HAPI server.
The ESA HAPI server requires authentication.
Users have to register with the ESA server
    and replace their username and password in the code below.

Notes
-----
https://swe.ssa.esa.int/sosmag
https://swe.ssa.esa.int/hapi
Data types available:
    1. (datatype='1m', esa_gk2a_sosmag_1m)
        Near-realtime Magnetic Field Data with 1-16Hz from SOSMAG
        on GEO-KOMPSAT-2A in geostationary orbit at 128.2E.
        'spase://SSA/NumericalData/GEO-KOMPSAT-2A/esa_gk2a_sosmag_1m'

    2. (default, datatype='', esa_gk2a_sosmag_recalib)
        Recalibrated L2 Magnetic Field Data with 1-16Hz from SOSMAG
        on GEO-KOMPSAT-2A in geostationary orbit at 128.2E.
        'spase://SSA/NumericalData/GEO-KOMPSAT-2A/esa_gk2a_sosmag_recalib'

Example
-------
from pytplot import tplot
from pyspedas import sosmag_load
t_ok, var_names = sosmag_load(trange=['2021-01-01 02:00','2021-01-01 03:00'])
tplot(var_names)

'''
import requests
import json
import numpy as np
from pytplot import store_data, options
from pyspedas import time_double, time_string


# Global parameters for the functions in this file.
# Users must replace username and password with their ESA credentials.
sosmag_parameters = {
    'username': 'spedas',  # Users must replace with their ESA username.
    'password': 'acQ4pG6u9Bh26v2',  # User ESA password.
    'portal_url': 'https://swe.ssa.esa.int/',
    'authenticate_url': 'https://sso.ssa.esa.int/am/json/authenticate',
    'sso_cookiename': 'iPlanetDirectoryPro',
    'print_errors': True,  # If true, prints error messages.
    'print_messages': False,  # If true, prints debugging/success messages.
}


def sosmag_get_auth_cookie():
    '''
    Authenticates a user against OpenAM.

    Uses the sosmag_parameters dictionary for the parameters needed.
    Returns whether authentication was successful and the obtained cookie.
    If an error occurs, the exception is caught and the error printed.

    Parameters
    ----------
    None.

    Returns
    -------
    success: bool
        True, if authentication was successful
    auth_cookie: str
        The obtained authentication cookie
    '''
    success = False
    auth_cookie = ''
    try:
        # Send a POST request to the authentication url
        response = requests.post(
            sosmag_parameters['authenticate_url'],
            headers={
                    'Content-Type': 'application/json',
                    'X-OpenAM-Username': sosmag_parameters['username'],
                    'X-OpenAM-Password': sosmag_parameters['password'],
                },
            data='{}')
        # form the response, extract the auth cookie and return it
        token_dict = json.loads(response.content)
        auth_cookie = token_dict['tokenId']
        success = True
    except Exception as exc:
        success = False
        auth_cookie = ''
        if sosmag_parameters['print_errors']:
            print(exc)
    finally:
        return success, auth_cookie


def sosmag_get_session(auth_cookie):
    '''
    Establishes a session with the ESA HAPI server.

    Uses the authentication cookie obtained from sosmag_get_auth_cookie().
    Returns whether a session was established successfully and if so,
        the obtained session cookies.
    If an error occurs, the exception is caught and the error is printed.

    Parameters
    ----------
    auth_cookie: str
        The obtained authentication cookie from sosmag_get_auth_cookie()

    Returns
    -------
    success: bool
        True, if authentication was successful
    jsession_id: str
        The obtained session cookie
    xsrf_token: str
        The obtained xsrf token
    '''
    success = False
    jsession_id = ''
    xsrf_token = ''
    try:
        # Try to access the HAPI/capabilities using the auth_cookie
        init_response = requests.get(
            sosmag_parameters['portal_url'] + '/hapi/capabilities',
            cookies={
                sosmag_parameters['sso_cookiename']: auth_cookie,
            }
        )
        # Extract the session cookies from the very first response from HAPI
        cookie_jar = init_response.history[0].cookies
        jsession_id = cookie_jar.get('JSESSIONID')
        xsrf_token = cookie_jar.get('XSRF-TOKEN')
        # Extract the consent url we are being requested to send our consent to
        # (in case we didn't consent yet)
        consent_url = init_response.url
        content = init_response.content

        # If we consented already, we should have received HAPI/capabilities.
        if '/hapi/capabilities' not in consent_url:
            # If not, we need to give our consent in the next step.
            # Send the consent along with all cookies to the consent url.
            consent_response = requests.post(
                consent_url,
                cookies={
                    sosmag_parameters['sso_cookiename']: auth_cookie,
                    'JSESSIONID': jsession_id,
                    'XSRF_TOKEN': xsrf_token,
                },
                data={
                    'decision': 'Allow',
                    'save_consent': 'on',
                }
            )
            content = consent_response.content
        # This will result in a redirect to the initial HAPI/capabilities.
        capabilities = json.loads(content)
        # The json output should be:
        #  2022/10/24: {'HAPI':'2.1.0','status':{'code':1200,'message':'OK'},'outputFormats':['csv','json']}
        #  previous: {'version':'2.1.0','status':{'code':1200,'message':'OK'},'outputFormats':['csv','json']}
        version = capabilities['HAPI']
        status = capabilities['status']
        # If the output is what we expect, return True and the session cookies
        if version != '' and status != {} and status['message'] == 'OK':
            success = True
        else:
            success = False
            jsession_id = ''
            xsrf_token = ''
    except Exception as exc:
        if sosmag_parameters['print_errors']:
            print(exc)
        success = False
        jsession_id = ''
        xsrf_token = ''

    return success, jsession_id, xsrf_token


def sosmag_get_capabilities(jsession_id, xsrf_token):
    '''
    Gets HAPI/capabilities from server.

    Uses the two cookies obtained using sosmag_get_session_cookies().
    If an error occurs, the exception is caught and the error printed.

    Parameters
    ----------
    jsession_id: str
        The obtained session cookie
    xsrf_token: str
        The obtained xsrf token

    Returns
    -------
    success: bool
        True, if function was successful
    capabilities: dict of str
        The obtained capabilities as a json dictionary of strings
    '''
    success = False
    capabilities = {}
    try:
        # Send a GET request to the HAPI/capabilities.
        test_response = requests.get(
            sosmag_parameters['portal_url'] + '/hapi/capabilities',
            cookies={
                'JSESSIONID': jsession_id,
                'XSRF_TOKEN': xsrf_token,
            }
        )
        # Extract the capabilities from the response.
        capabilities = json.loads(test_response.content)
        version = capabilities['HAPI']
        status = capabilities['status']
        # If the capabilities are as expected, return True.
        if version != '' and status != {} and status['message'] == 'OK':
            success = True
        else:
            success = False
            capabilities = {}
    except Exception as exc:
        if sosmag_parameters['print_errors']:
            print(exc)
        success = False
        capabilities = {}

    return success, capabilities


def sosmag_get_data(
    jsession_id, xsrf_token,  # HAPI session cookies.
    datatype='',   # Data type, either '' (recalibrated) or '1m' (real time)
    timemin='2022-01-31T01:00:00.000Z',  # Start time
    timemax='2022-01-31T01:10:00.000Z',  # End time
):
    '''
    Gets either HAPI/data or HAPI/info from server.

    Uses the two cookies obtained from sosmag_get_session_cookies().
    If an error occurs, the exception is caught and the error printed.

    Parameters
    ----------
    jsession_id: str
        The obtained session cookie.
    xsrf_token: str
        The obtained xsrf token cookie.
    datatype: str
        Data type, either '' (recalibrated) or '1m' (1 minute, real time).
    timemin: str
        Start time, for example: '2022-01-31T01:00:00.000Z'.
    timemax: str
        End time, for example: '2022-01-31T01:10:00.000Z'.

    Returns
    -------
    success: bool
        True, if function was successful.
    data: dict of str
        The obtained data as a json dictionary of strings.
    parameters: dict of str
        The obtained parameters for the data as a json dictionary of strings.
    description: str
        The obtained description of data.
    '''
    success = False
    data = {}
    parameters = {}
    description = ''
    dataid = 'spase://SSA/NumericalData/GEO-KOMPSAT-2A/'
    if datatype == '1m':
        dataid = dataid + 'esa_gk2a_sosmag_1m'
    else:
        dataid = dataid + 'esa_gk2a_sosmag_recalib'

    # This query string can also be used in a browser.
    hquery = ('data?id=' + dataid +
              '&time.min=' + timemin +
              '&time.max=' + timemax +
              '&format=json')
    fullurl = sosmag_parameters['portal_url'] + '/hapi/' + hquery
    if sosmag_parameters['print_messages']:
        print(fullurl)

    try:
        # send a GET request to the HAPI server
        test_response = requests.get(
            fullurl,
            cookies={
                'JSESSIONID': jsession_id,
                'XSRF_TOKEN': xsrf_token,
            }
        )
        # extract the content from the response
        dat = json.loads(test_response.content)
        data = dat['data']
        parameters = dat['parameters']
        description = dat['description']
        status = dat['status']

        # if the status are as expected, return True along with the data dict
        if data != '' and status != {} and status['message'] == 'OK':
            success = True
        else:
            success = False
            data = {}
            parameters = {}
            description = ''
    except Exception as exc:
        if sosmag_parameters['print_errors']:
            print(exc)
        success = False
        data = {}
        parameters = {}
        description = ''

    return success, data, parameters, description


def sosmag_to_tplot(
    data,  # Data, array with 14 fields
    parameters,  # Description of data, array with 14 fields
    desc,   # Single string description of the data set
    datatype='',  # Either '' (recalibrated) or '1m' (real time)
    prefix='',  # To be added as prefix of tplot variable
    suffix=''  # To be added as suffix of tplot variable
):
    '''
    Saves SOSMAG data as tplot variables.

    Creates three tplot variables:
        1. Magnetic Field B in GSE coordinates, sosmag_b_gse.
        2. Magnetic Field B in HPEN coordinates, sosmag_b_hpen.
        3. Spacecraft Position in GSE, sosmag_pos.

    Parameters
    ----------
    data: dict of str
        The obtained data as a json dictionary of strings.
    parameters: dict of str
        The obtained parameters for the data as a json dictionary of strings.
    description: str
        The obtained description for this set of data.
    datatype: str
        Data type can be either '' (recalibrated) or '1m' (1 minute real time).
    prefix: str
        Prefix for tplot names. Default is ''.
    suffix: str
        Suffix for tplot names. Default is ''.

    Returns
    -------
    success: bool
        True, if function was successful.
    var_names: list of str
        Names of tplot variables created.
    '''
    success = False
    var_names = []
    try:
        # Construct tplot variable names.
        if datatype == '':
            pre = prefix + 'sosmag'
        else:
            pre = prefix + 'sosmag_' + datatype

        # Get data
        d = np.array(data)  # data
        p = np.array(parameters)  # parameters

        if len(d.shape) != 2 or len(p.shape) != 1 or d.shape[1] != 14:
            print('SOSMAG data has wrong shape. Abort.')
            return False

        # Time is the 0th field
        td = np.array(time_double(d[:, 0]))

        # Magnetic field in GSE 'b_gse_x', 'b_gse_y', 'b_gse_z'
        # Data fields: [2, 3, 4]
        yd = np.array(d[:, 2:5], dtype=float)
        var_name0 = pre + '_b_gse' + suffix
        pnames = [p[2]['name'], p[3]['name'], p[4]['name']]
        pd = 'Magnetic Field B in GSE coordinates'
        attr_dict = {'description': pd}
        store_data(var_name0, data={'x': td, 'y': yd}, attr_dict=attr_dict)
        options(var_name0, 'legend_names', pnames)
        options(var_name0, 'ysubtitle', '[nT]')
        options(var_name0, 'coord_sys', 'gse')
        options(var_name0, 'description', desc)

        # Magnetic field in HPEN 'b_hpen_x', 'b_hpen_y', 'b_hpen_z'
        # Data fields: [5, 6, 7]
        yd = np.array(d[:, 5:8], dtype=float)
        var_name1 = pre + '_b_hpen' + suffix
        pnames = [p[5]['name'], p[6]['name'], p[7]['name']]
        pd = 'Magnetic Field B in HPEN coordinates'
        attr_dict = {'description': pd}
        store_data(var_name1, data={'x': td, 'y': yd}, attr_dict=attr_dict)
        options(var_name1, 'legend_names', pnames)
        options(var_name1, 'ysubtitle', '[nT]')
        options(var_name1, 'coord_sys', 'hpen')
        options(var_name1, 'description', desc)

        # Spacecraft Position in GSE 'position_x', 'position_y', 'position_z'
        # Data fields: [8, 9, 10]
        yd = np.array(d[:, 8:11], dtype=float)
        var_name2 = pre + '_position' + suffix
        pnames = [p[8]['name'], p[9]['name'], p[10]['name']]
        pd = 'Spacecraft Position in GSE'
        attr_dict = {'description': pd}
        store_data(var_name2, data={'x': td, 'y': yd}, attr_dict=attr_dict)
        options(var_name2, 'legend_names', pnames)
        options(var_name2, 'ysubtitle', '[km]')
        options(var_name2, 'coord_sys', 'gse')
        options(var_name2, 'description', desc)

        # Print the variable names created
        var_names = [var_name0, var_name1, var_name2]
        print('Tplot variables created!')
        print(var_names)
        success = True
    except Exception as exc:
        if sosmag_parameters['print_errors']:
            print(exc)
        success = False

    return success, var_names


def sosmag_load(
    trange=['2022-01-31T01:00:00.000Z',
            '2022-01-31T01:01:00.000Z'],  # Default 10 minutes of data
    datatype='',  # Either '' (recalibrated) or '1m' (real time)
    prefix='',  # Prefix for tplot names
    suffix='',  # Suffix for tplot names
):
    '''
    This function loads data from SOSMAG.

    Gets data from the ESA web server as a json string.
    Creates three tplot variables:
        1. Magnetic Field B in GSE coordinates, sosmag_b_gse
        2. Magnetic Field B in HPEN coordinates, sosmag_b_hpen
        3. Spacecraft Position in GSE, sosmag_pos

    Parameters
    ----------
    trange: list of str
        Start and end times.
        Many time formats are supported, see function time_double().
    datatype: str
        Either '' (recalibrated) or '1m' (real time).
    prefix: str
        Prefix for tplot names. Default is ''.
    suffix: str
        Suffix for tplot names. Default is ''.

    Returns
    -------
    success: bool
        True, if function was successful.
    var_names: list of str
        Names of tplot variables created.

    Notes
    -----
    Link that can be tested on a browser:
    https://swe.ssa.esa.int/hapi/data?
    id=spase://SSA/NumericalData/GEO-KOMPSAT-2A/esa_gk2a_sosmag_recalib
    &time.min=2021-01-31T01:00:00.000Z
    &time.max=2021-01-31T01:01:00.000Z&format=json
    '''
    success = False
    var_names = []
    # Make sure that time is in the correct format for the ESA server.
    trange = time_string(time_double(trange), fmt='%Y-%m-%dT%H:%M:%S.000Z')
    if sosmag_parameters['print_messages']:
        print('trange', trange)

    # Authenticate with the HAPI server and receive a cookie.
    authenticated, auth_cookie = sosmag_get_auth_cookie()
    if authenticated:
        if sosmag_parameters['print_messages']:
            print('SOSMAG authentication successful.')
    else:
        print('SOSMAG authentication failed. \
        Please check username and password in the file sosmag:load.py.')
        return

    # Set a session with the server.
    capok = False
    if authenticated:
        sessionοκ, jsession_id, xsrf_token = sosmag_get_session(auth_cookie)
        if sessionοκ:
            # Session successfully established. Obtained session cookies.
            # Testing session cookies on hapi/capabilities.
            if sosmag_parameters['print_messages']:
                print('SOSMAG session established successfully.')
            capok, capabilities = sosmag_get_capabilities(jsession_id,
                                                          xsrf_token)
            if capok:
                if sosmag_parameters['print_messages']:
                    print('SOSMAG HAPI server capabilities: ')
                    print(capabilities)
            else:
                print('Problem communicating with server. Aborting.')
                return
        else:
            print('SOSMAG session could not be established. Aborting.')
            return

    # Get data.
    dataok = False
    if capok:
        dataok, data, parameters, desc = sosmag_get_data(
            jsession_id, xsrf_token, datatype=datatype,
            timemin=trange[0], timemax=trange[1])
        if dataok:
            tplotok, var_names = sosmag_to_tplot(
                data, parameters, desc, datatype=datatype,
                prefix=prefix, suffix=suffix)
            if tplotok:
                if sosmag_parameters['print_messages']:
                    print('Data was loaded.')
                success = True
            else:
                print('Could load data into pytplot. Aborting.')
                return
        else:
            print('Could not get any data. Aborting.')
            return

    return success, var_names



import numpy as np

from astropy.coordinates import cartesian_to_spherical

from .get_mepe_az_dir_in_sga import get_mepe_az_dir_in_sga


def get_mepe_flux_angle_in_sga(looking_dir=False):

    fluxdir = not looking_dir
    sgajdir = get_mepe_az_dir_in_sga(fluxdir=fluxdir)

    spherical_data = cartesian_to_spherical(
        x=sgajdir[0],
        y=sgajdir[1],
        z=sgajdir[2]
    )

    rtod = 180. / np.pi
    elev_array = spherical_data[1].value * rtod
    phi_array = spherical_data[2].value * rtod

    anglarr = np.zeros(shape=(2, 3, 16))  # ;[ elev/phi, min/cnt/max, apd_no ]

    anglarr[0] = elev_array
    anglarr[1] = phi_array

    return anglarr
